# Copyright contributors to the TSFM project
#
"""Hugging Face Pipeline for Time Series Tasks"""

import inspect
from collections import defaultdict
from typing import Any, Dict, List, Union

import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader
from transformers import PreTrainedModel
from transformers.data.data_collator import default_data_collator
from transformers.pipelines.base import (
    GenericTensor,
    Pipeline,
    build_pipeline_init_args,
)
from transformers.trainer_utils import RemoveColumnsCollator
from transformers.utils import add_end_docstrings, logging

from .dataset import ForecastDFDataset
from .time_series_preprocessor import create_timestamps, extend_time_series


# Eventually we should support all time series models
# MODEL_FOR_TIME_SERIES_FORECASTING_MAPPING_NAMES = OrderedDict(
#     [
#         # Model for Time Series Forecasting
#         ("PatchTST", "PatchTSTForPrediction"),
#         ("TST", "TimeSeriesTransformerForPrediction"),
#     ]
# )


logger = logging.get_logger(__name__)


class TimeSeriesPipeline(Pipeline):
    def run_single(self, inputs, preprocess_params, forward_params, postprocess_params):
        """Replaces base `run_single` method which does batching during inference. This is needed to support
        large inference requests.

        Args:
            inputs (_type_): _description_
            preprocess_params (_type_): _description_
            forward_params (_type_): _description_
            postprocess_params (_type_): _description_

        Returns:
            _type_: _description_
        """
        # our preprocess returns a dataset
        dataset = self.preprocess(inputs, **preprocess_params)["dataset"]
        model_output_key = getattr(self, "_model_output_key", None)
        copy_dataset_keys = getattr(self, "_copy_dataset_keys", True)

        batch_size = forward_params["batch_size"]
        num_workers = forward_params["num_workers"]
        signature = inspect.signature(self.model.forward)
        signature_columns = list(signature.parameters.keys())

        # if len(dataset) < batch_size:
        # build a dataloader
        # collate_fn = no_collate_fn if batch_size == 1 else pad_collate_fn(self.tokenizer, feature_extractor)

        remove_columns_collator = RemoveColumnsCollator(
            data_collator=default_data_collator,
            signature_columns=signature_columns,
            logger=None,
            description=None,
            model_name=None,
        )
        dataloader = DataLoader(
            dataset, batch_size=batch_size, num_workers=num_workers, collate_fn=remove_columns_collator, shuffle=False
        )

        # iterate over dataloader
        it = iter(dataloader)
        accumulator = []
        while (batch := next(it, None)) is not None:
            item = self.forward(batch, **forward_params)
            if not model_output_key:
                model_output_key = "prediction_outputs" if "prediction_outputs" in item.keys() else "prediction_logits"
            accumulator.append(item[model_output_key])

        if copy_dataset_keys:
            # collect all ouputs needed for post processing
            model_outputs = defaultdict(list)
            items = list(dataset[0].items())
            for r in dataset:
                for k, v in items:
                    model_outputs[k].append(r[k])

            for k, v in items:
                if isinstance(v, torch.Tensor):
                    model_outputs[k] = torch.stack(model_outputs[k])

            # without shuffling in the dataloader above, we assume that order is preserved
            # otherwise we need to incorporate sequence id somewhere and do a proper join
            model_outputs["prediction_outputs"] = torch.cat(accumulator, axis=0)
        else:
            model_outputs = accumulator

        # call postprocess
        outputs = self.postprocess(model_outputs, **postprocess_params)

        return outputs


@add_end_docstrings(
    build_pipeline_init_args(has_tokenizer=False, has_feature_extractor=True, has_image_processor=False)
)
class TimeSeriesForecastingPipeline(TimeSeriesPipeline):
    """
    Time Series Forecasting using HF time series forecasting models. This pipeline consumes a `pandas.DataFrame`
    containing the time series data and produces a new `pandas.DataFrame` containing the forecasts.

    """

    def __init__(
        self,
        model: Union["PreTrainedModel"],
        *args,
        explode_forecasts: bool = False,
        inverse_scale_outputs: bool = True,
        add_known_ground_truth: bool = True,
        **kwargs,
    ):
        kwargs["explode_forecasts"] = explode_forecasts
        kwargs["inverse_scale_outputs"] = inverse_scale_outputs
        kwargs["add_known_ground_truth"] = add_known_ground_truth

        # autopopulate from feature extractor and model
        if "feature_extractor" in kwargs:
            for p in [
                "id_columns",
                "timestamp_column",
                "target_columns",
                "observable_columns",
                "control_columns",
                "conditional_columns",
                "categorical_columns",
                "static_categorical_columns",
                "freq",
            ]:
                if p not in kwargs:
                    kwargs[p] = getattr(kwargs["feature_extractor"], p)

        if "context_length" not in kwargs:
            kwargs["context_length"] = model.config.context_length

        if "prediction_length" not in kwargs:
            kwargs["prediction_length"] = model.config.prediction_length

        # check if we need to use the frequency token, get token if needed
        use_frequency_token = getattr(model.config, "resolution_prefix_tuning", False)

        if use_frequency_token and ("feature_extractor" not in kwargs) and "freq" in kwargs:
            raise ValueError(
                "Passing `freq` without a `feature_extractor` is not supported when the model requires a `frequency_token`."
            )

        if use_frequency_token:
            if "feature_extractor" in kwargs:
                kwargs["frequency_token"] = kwargs["feature_extractor"].get_frequency_token(kwargs["freq"])
        else:
            kwargs["frequency_token"] = None

        super().__init__(model, *args, **kwargs)

        if self.framework == "tf":
            raise ValueError(f"The {self.__class__} is only available in PyTorch.")

        # self.check_model_type(MODEL_FOR_TIME_SERIES_FORECASTING_MAPPING)

    def _sanitize_parameters(
        self,
        **kwargs,
    ):
        """Assigns parameters to the different steps of the process. If context_length and prediction_length
        are not provided they are taken from the model config.

        For expected parameters see the call method below.
        """

        preprocess_kwargs = {}
        postprocess_kwargs = {}

        preprocess_params = [
            "prediction_length",
            "context_length",
            "frequency_token",
            "id_columns",
            "timestamp_column",
            "target_columns",
            "observable_columns",
            "control_columns",
            "conditional_columns",
            "categorical_columns",
            "static_categorical_columns",
            "future_time_series",
            "impute_method",
        ]
        postprocess_params = [
            "prediction_length",
            "context_length",
            "id_columns",
            "timestamp_column",
            "target_columns",
            "observable_columns",
            "control_columns",
            "conditional_columns",
            "categorical_columns",
            "static_categorical_columns",
            "freq",
            "explode_forecasts",
            "inverse_scale_outputs",
            "add_known_ground_truth",
        ]

        for c in preprocess_params:
            if c in kwargs:
                preprocess_kwargs[c] = kwargs[c]

        for c in postprocess_params:
            if c in kwargs:
                postprocess_kwargs[c] = kwargs[c]

        # same logic as HF Pipeline
        batch_size = kwargs.get("batch_size", self._batch_size)
        if batch_size is None:
            if self._batch_size is None:
                batch_size = 1
            else:
                batch_size = self._batch_size

        num_workers = kwargs.get("num_workers", self._num_workers)
        if num_workers is None:
            if self._num_workers is None:
                num_workers = 0
            else:
                num_workers = self._num_workers

        forward_kwargs = {"batch_size": batch_size, "num_workers": num_workers}

        # if "id_columns" in kwargs:
        #     preprocess_kwargs["id_columns"] = kwargs["id_columns"]
        #     postprocess_kwargs["id_columns"] = kwargs["id_columns"]
        # if "timestamp_column" in kwargs:
        #     preprocess_kwargs["timestamp_column"] = kwargs["timestamp_column"]
        #     postprocess_kwargs["timestamp_column"] = kwargs["timestamp_column"]
        # if "input_columns" in kwargs:
        #     preprocess_kwargs["input_columns"] = kwargs["input_columns"]
        #     postprocess_kwargs["input_columns"] = kwargs["input_columns"]
        # if "output_columns" in kwargs:
        #     preprocess_kwargs["output_columns"] = kwargs["output_columns"]
        #     postprocess_kwargs["output_columns"] = kwargs["output_columns"]
        # elif "input_columns" in kwargs:
        #     preprocess_kwargs["output_columns"] = kwargs["input_columns"]
        #     postprocess_kwargs["output_columns"] = kwargs["input_columns"]

        return preprocess_kwargs, forward_kwargs, postprocess_kwargs

    def __call__(
        self,
        time_series: Union["pd.DataFrame", str],
        **kwargs,
    ):
        """Main method of the forecasting pipeline. Takes the input time series data (in tabular format) and
        produces predictions.

        Args:
            time_series (Union[&quot;pandas.DataFrame&quot;, str]): A pandas dataframe containing the time series on
                which to perform inference.

        Keyword arguments:
            future_time_series (Union[&quot;pandas.DataFrame&quot;, str]): A pandas dataframe containing future values,
                i.e., exogenous or supporting features which are known in advance.

            feature_extractor (TimeSeriesPreprocessor): A time series preprpocessor object that specifies how the time
                series should be prepared. If this is provided, any of the other options below will be automatically
                populated from this instance.

            timestamp_column (str): The name of the column containing the timestamp of the time series.

            id_columns (List[str]): List of column names which identify different time series in a multi-time series input.

            target_columns (List[str]): List of column names which identify the target channels in the input, these are the
                columns that will be forecasted.

            observable_columns (List[str]): List of column names which identify the observable channels in the input.
                Observable channels are channels which we have knowledge about in the past and future. For example, weather
                conditions such as temperature or precipitation may be known or estimated in the future, but cannot be
                changed.

            control_columns (List[str]): List of column names which identify the control channels in the input. Control
                channels are similar to observable channels, except that future values may be controlled. For example, discount
                percentage of a particular product is known and controllable in the future.

            conditional_columns (List[str]): List of column names which identify the conditional channels in the input.
                Conditional channels are channels which we know in the past, but do not know in the future.

            categorical_columns (List[str]): List of column names which identify time-varying categorical-valued channels in the input.

            static_categorical_columns (List[str]): List of column names which identify categorical-valued channels in the input
                which are fixed over time.

            freq (str): A freqency indicator for the given `timestamp_column`. See
                https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#period-aliases for a description of the
                allowed values. If not provided, we will attempt to infer it from the data. If not provided, frequency will be
                inferred from `timestamp_column`.

            prediction_length (int): The length of the desired forecast. Currently, this value must not exceed the maximum value
                suported by the model. If not specified, the maximum value supported by the model is used.

            context_length (int): Specifies the length of the context windows extracted from the historical data for feeding into
                the model.

            explode_forecasts (bool): If true, forecasts are returned one value per row of the pandas dataframe. If false, the
                forecast over the prediction length will be contained as a list in a single row of the pandas dataframe.

            inverse_scale_outputs (bool): If true and a valid feature extractor is provided, the outputs will be inverse scaled.

            add_known_ground_truth (bool): If True add columns containing the ground truth data. Prediction columns will have a
                suffix of "_prediction". Default True. If false, on columns containing predictions are produced, no suffix is
                added.

        Return (pandas dataframe):
            A new pandas dataframe containing the forecasts. Each row will contain the id, timestamp, the original
            input feature values and the output forecast for each input column. The output forecast is a list containing
            all the values over the prediction horizon.

        """

        return super().__call__(time_series, **kwargs)

    def preprocess(self, time_series, **kwargs) -> Dict[str, Union[GenericTensor, List[Any]]]:
        """Preprocess step
        Load the data, if not already loaded, and then generate a pytorch dataset.
        """

        prediction_length = kwargs.get("prediction_length")
        timestamp_column = kwargs.get("timestamp_column")
        id_columns = kwargs.get("id_columns")
        target_columns = kwargs.get("target_columns")
        # context_length = kwargs.get("context_length")

        # use the feature extractor here

        if isinstance(time_series, str):
            time_series = pd.read_csv(
                time_series,
                parse_dates=[timestamp_column],
            )

        future_time_series = kwargs.pop("future_time_series", None)

        if self.feature_extractor:
            time_series = self.feature_extractor.preprocess(time_series)

        if future_time_series is not None:
            if isinstance(future_time_series, str):
                future_time_series = pd.read_csv(
                    future_time_series,
                    parse_dates=[timestamp_column],
                )
            elif isinstance(future_time_series, pd.DataFrame):
                # do we need to check the timestamp column?
                pass
            else:
                raise ValueError(f"`future_time_series` of type {type(future_time_series)} is not supported.")

            # stack the time series
            for c in future_time_series.columns:
                if c not in time_series.columns:
                    raise ValueError(f"Future time series input contains an unknown column {c}.")

            if id_columns:
                id_count = time_series[id_columns].drop_duplicates().shape[0]
            else:
                id_count = 1

            if future_time_series.shape[0] != prediction_length * id_count:
                raise ValueError(
                    f"If provided, `future_time_series` data should cover the prediction length for each of the time series in the test dataset. Received data of length {future_time_series.shape[0]} but expected {prediction_length * id_count}"
                )

            if self.feature_extractor:
                # future data needs some values for targets, but they are unused
                future_time_series[target_columns] = 0
                future_time_series = self.feature_extractor.preprocess(future_time_series)
                # future_time_series = future_time_series.drop(columns=target_columns)
                future_time_series[target_columns] = np.nan

            time_series = pd.concat((time_series, future_time_series), axis=0, ignore_index=True)
        else:
            # no additional exogenous data provided, extend with empty periods
            time_series = extend_time_series(
                time_series=time_series,
                freq=self.feature_extractor.freq if self.feature_extractor else None,
                timestamp_column=timestamp_column,
                grouping_columns=id_columns,
                periods=prediction_length,
            )

        # use forecasting dataset to do the preprocessing
        dataset = ForecastDFDataset(
            time_series,
            **kwargs,
        )

        return {"dataset": dataset}

    def _forward(self, model_inputs, **kwargs):
        """Forward step
        Responsible for taking pre-processed dictionary of tensors and passing it to
        the model. Aligns model parameters with the proper input parameters. Only passes
        the needed parameters from the dictionary to the model, but adds them back to the
        ouput for the next step.

        The keys in model_outputs are governed by the underlying model combined with any
        original input keys.
        """

        model_outputs = self.model(**model_inputs)

        return model_outputs

    def postprocess(self, input, **kwargs):
        """Postprocess step
        Takes the dictionary of outputs from the previous step and converts to a more user
        readable pandas format.

        If the explode forecasts option is True, then individual forecasts are expanded as multiple
        rows in the dataframe. This should only be used when producing a single forecast (i.e., unexploded
        result is one row per ID).
        """
        out = {}

        model_output_key = "prediction_outputs"  #  if "prediction_outputs" in input.keys() else "prediction_logits"

        # name the predictions of target columns
        # outputs should only have size equal to target columns

        # only allow adding ground truth when not exploding
        add_known_ground_truth = kwargs["add_known_ground_truth"] if not kwargs["explode_forecasts"] else False

        prediction_columns = []
        for i, c in enumerate(kwargs["target_columns"]):
            prediction_columns.append(f"{c}_prediction" if add_known_ground_truth else c)
            out[prediction_columns[-1]] = input[model_output_key][:, :, i].detach().cpu().numpy().tolist()
        # provide the ground truth values for the targets
        # when future is unknown, we will have augmented the provided dataframe with NaN values to cover the future
        if add_known_ground_truth:
            for i, c in enumerate(kwargs["target_columns"]):
                ground_truth = input["future_values"][:, :, i].detach().cpu().numpy()
                missing = ~input["future_observed_mask"][:, :, i].detach().cpu().numpy()
                ground_truth[missing] = np.nan
                out[c] = ground_truth.tolist()

        if "timestamp_column" in kwargs:
            out[kwargs["timestamp_column"]] = input["timestamp"]
        # copy ids
        for i, c in enumerate(kwargs["id_columns"]):
            out[c] = [elem[i] for elem in input["id"]]
        out = pd.DataFrame(out)

        if kwargs["explode_forecasts"]:
            # we made only one forecast per time series, explode results
            # explode == expand the lists in the dataframe
            out_explode = []
            for _, row in out.iterrows():
                l = len(row[prediction_columns[0]])
                tmp = {}
                if "timestamp_column" in kwargs:
                    tmp[kwargs["timestamp_column"]] = create_timestamps(
                        row[kwargs["timestamp_column"]], freq=kwargs["freq"], periods=l
                    )  # expand timestamps
                if "id_columns" in kwargs:
                    for c in kwargs["id_columns"]:
                        tmp[c] = row[c]
                for p in prediction_columns:
                    tmp[p] = row[p]

                out_explode.append(pd.DataFrame(tmp))

            out = pd.concat(out_explode)

        # reorder columns
        cols = out.columns.to_list()
        cols_ordered = []
        if "timestamp_column" in kwargs:
            cols_ordered.append(kwargs["timestamp_column"])
        if "id_columns" in kwargs:
            cols_ordered.extend(kwargs["id_columns"])
        cols_ordered.extend([c for c in cols if c not in cols_ordered])

        out = out[cols_ordered]

        # inverse scale if we have a feature extractor
        if self.feature_extractor is not None and kwargs["inverse_scale_outputs"]:
            out = self.feature_extractor.inverse_scale_targets(out)
            if add_known_ground_truth:
                out = self.feature_extractor.inverse_scale_targets(out, suffix="_prediction")

        return out



# Copyright contributors to the TSFM project
#

import logging
import os
from pathlib import Path
from typing import TYPE_CHECKING

# Check the dependencies satisfy the minimal versions required.
from transformers.utils import _LazyModule

from tsfm_public.toolkit.hf_util import register_config

from .version import __version__, __version_tuple__


TSFM_PYTHON_LOGGING_LEVEL = os.getenv("TSFM_PYTHON_LOGGING_LEVEL", "INFO")

LevelNamesMapping = {
    "INFO": logging.INFO,
    "WARN": logging.WARN,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
    "DEBUG": logging.DEBUG,
    "FATAL": logging.FATAL,
}

TSFM_PYTHON_LOGGING_LEVEL = (
    logging.getLevelNamesMapping()[TSFM_PYTHON_LOGGING_LEVEL]
    if hasattr(logging, "getLevelNamesMapping")
    else LevelNamesMapping[TSFM_PYTHON_LOGGING_LEVEL]
)
TSFM_PYTHON_LOGGING_FORMAT = os.getenv(
    "TSFM_PYTHON_LOGGING_FORMAT",
    "%(levelname)s:p-%(process)d:t-%(thread)d:%(filename)s:%(funcName)s:%(message)s",
)

logging.basicConfig(
    format=TSFM_PYTHON_LOGGING_FORMAT,
    level=TSFM_PYTHON_LOGGING_LEVEL,
)

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

# Base objects, independent of any specific backend
_import_structure = {
    "models": [],
    "models.tinytimemixer": ["TINYTIMEMIXER_PRETRAINED_CONFIG_ARCHIVE_MAP", "TinyTimeMixerConfig"],
    "models.tspulse": ["TINYTIMEMIXER_PRETRAINED_CONFIG_ARCHIVE_MAP", "TSPulseConfig"],
    "toolkit": [
        "TimeSeriesPreprocessor",
        "TimeSeriesForecastingPipeline",
        "ClassificationDFDataset",
        "ForecastDFDataset",
        "PretrainDFDataset",
        "RegressionDFDataset",
        "get_datasets",
        "load_dataset",
        "get_model",
        "count_parameters",
        "TrackingCallback",
        "count_parameters",
        "optimal_lr_finder",
        "ClassificationDFDataset",
        "TimeSeriesImputationPipeline",
        "TimeSeriesClassificationPipeline",
        "TimeSeriesAnomalyDetectionPipeline",
        "TimeSeriesPreprocessor",
        "TimeSeriesClassificationPreprocessor",
    ],
}


# PyTorch-backed objects
_import_structure["models.tinytimemixer"].extend(
    [
        "TINYTIMEMIXER_PRETRAINED_MODEL_ARCHIVE_LIST",
        "TinyTimeMixerPreTrainedModel",
        "TinyTimeMixerModel",
        "TinyTimeMixerForMaskedPrediction",
        "TinyTimeMixerForPrediction",
    ]
)
_import_structure["models.tspulse"].extend(
    [
        "TSPULSE_PRETRAINED_MODEL_ARCHIVE_LIST",
        "TSPulsePreTrainedModel",
        "TSPulseModel",
        "TSPulseForReconstruction",
        "TSPulseForClassification",
    ]
)

# Direct imports for type-checking
if TYPE_CHECKING:
    from .models.tinytimemixer import (
        TINYTIMEMIXER_PRETRAINED_CONFIG_ARCHIVE_MAP,
        TINYTIMEMIXER_PRETRAINED_MODEL_ARCHIVE_LIST,
        TinyTimeMixerConfig,
        TinyTimeMixerForMaskedPrediction,
        TinyTimeMixerForPrediction,
        TinyTimeMixerModel,
        TinyTimeMixerPreTrainedModel,
    )
    from .models.tspulse import (
        TSPULSE_PRETRAINED_CONFIG_ARCHIVE_MAP,
        TSPULSE_PRETRAINED_MODEL_ARCHIVE_LIST,
        TSPulseConfig,
        TSPulseForClassification,
        TSPulseForReconstruction,
        TSPulseModel,
        TSPulsePreTrainedModel,
    )
    from .toolkit import (
        ClassificationDFDataset,
        ForecastDFDataset,
        PretrainDFDataset,
        RegressionDFDataset,
        TimeSeriesAnomalyDetectionPipeline,
        TimeSeriesClassificationPipeline,
        TimeSeriesClassificationPreprocessor,
        TimeSeriesForecastingPipeline,
        TimeSeriesImputationPipeline,
        TimeSeriesPreprocessor,
        TrackingCallback,
        count_parameters,
        get_datasets,
        get_model,
        load_dataset,
        optimal_lr_finder,
    )
else:
    import sys

    sys.modules[__name__] = _LazyModule(
        __name__,
        globals()["__file__"],
        _import_structure,
        module_spec=__spec__,
        extra_objects={"__version__": __version__},
    )


# register local models now
register_config(model_type="tinytimemixer", model_config_name="TinyTimeMixerConfig", module_path="tsfm_public")
register_config(model_type="tspulse", model_config_name="TSPulseConfig", module_path="tsfm_public")



# Copyright contributors to the TSFM project
#
"""Utilities for the demo app"""

import copy
import logging
import os
from typing import List, Type

import pandas as pd
import streamlit as st
import transformers
from plotly.graph_objs import graph_objs
from transformers import AutoConfig
from tsevaluate.multivalue_timeseries_evaluator import CrossTimeSeriesEvaluator

# Local
from tsfm_public.toolkit.time_series_forecasting_pipeline import (
    TimeSeriesForecastingPipeline,
)
from tsfm_public.toolkit.time_series_preprocessor import TimeSeriesPreprocessor
from tsfm_public.toolkit.util import select_by_index
from tsfm_public.toolkit.visualization import plot_ts_forecasting


# A dictionary containing datasets mapped to their location
# Note that keys of this dictionary get rendered in the UI

LOGGER = logging.getLogger(__file__)


@st.cache_data(persist="disk")
def get_performance(metrics: List[str] = None, **kwargs) -> pd.DataFrame:
    """Return a dataframe of performance results

    Args:
        metainfo (dict): _description_

    Returns:
        pd.DataFrame: _description_
    """

    print("perf: ", kwargs)
    # we want to remove some parameters that
    # do not get used within our forecast method
    # because passing them will break the cache from
    # call to call
    # cache_breakers = ["channel"]
    # cachable_args = copy.deepcopy(kwargs)
    # _ = [cachable_args.pop(arg) for arg in cache_breakers]
    timestamp_column = kwargs["timestamp_column"]
    id_columns = kwargs["id_columns"]
    forecast_columns = kwargs["forecast_columns"]

    _, forecasts = forecast(**kwargs)

    labels_ = forecasts[id_columns + [timestamp_column] + forecast_columns]
    forecasts_ = forecasts.drop(columns=forecast_columns)

    tseval = CrossTimeSeriesEvaluator(
        timestamp_column=timestamp_column,
        prediction_columns=[f"{c}_prediction" for c in forecast_columns],
        label_columns=forecast_columns,
        metrics_spec=metrics,
        multioutput="raw",
    )
    df_eval = tseval.evaluate(labels_, forecasts_)

    # df_eval.columns = [c.split("_")[1] for c in df_eval.columns]

    dfm = df_eval.melt()

    dfm["channel"] = dfm["variable"].apply(lambda x: x.split("_")[0].upper())
    dfm["metric"] = dfm["variable"].apply(lambda x: x.split("_")[-1].upper())
    dfm = dfm.pivot(columns=["metric"], index=["channel"], values=["value"])
    dfm.columns = dfm.columns.droplevel(0)

    tseval = CrossTimeSeriesEvaluator(
        timestamp_column=timestamp_column,
        prediction_columns=[f"{c}_prediction" for c in forecast_columns],
        label_columns=forecast_columns,
        metrics_spec=metrics,
        multioutput="uniform_average",
    )
    df_eval = tseval.evaluate(labels_, forecasts_)

    df_eval.columns = [c.split("_")[1].upper() for c in df_eval.columns]
    df_eval.index = ["Average"]

    return pd.concat([dfm, df_eval], axis=0)


@st.cache_data(persist="disk")
def csv_to_df(metainfo: dict) -> pd.DataFrame:
    """Return a pandas dataframe for a given entry in DATASETS metainfo.

    Args:
        metainfo (dict): meta information about the dataset. This is the values in the key->value
        mapping in the DATASETS dictionary.

    Returns:
        pd.DataFrame: _description_
    """
    return pd.read_csv(metainfo["uri"], parse_dates=[metainfo["timestamp_column"]])


def get_model_class(model_path: str) -> Type:
    conf = AutoConfig.from_pretrained(model_path)
    model_class = getattr(transformers, conf.architectures[0])

    return model_class


def get_model_path(**kwargs) -> str:
    if kwargs["approach"] == "zero_shot":
        model_path = os.path.join(kwargs["pretrained_model_path"], "pretrain")
    else:
        model_path = os.path.join(
            kwargs["pretrained_model_path"],
            "transfer",
            kwargs["dataset_name"],
            "model",
            kwargs["approach"],
        )
    return model_path


def get_preprocessor_path(**kwargs) -> str:
    return os.path.join(
        kwargs["pretrained_model_path"],
        "transfer",
        kwargs["dataset_name"],
        "preprocessor",
    )


@st.cache_data(persist="disk")
def forecast(**kwargs) -> pd.DataFrame:
    LOGGER.debug("in forecast with:")
    LOGGER.debug(kwargs)

    timestamp_column = kwargs["timestamp_column"]
    id_columns = kwargs["id_columns"]
    forecast_columns = kwargs["forecast_columns"]
    # finetuned_model_path = kwargs["finetuned_model_path"]
    # pretrained_model_path = kwargs["pretrained_model_path"]

    model_path = get_model_path(**kwargs)
    prep_path = get_preprocessor_path(**kwargs)

    model_class = get_model_class(model_path)
    model = model_class.from_pretrained(model_path, num_input_channels=len(forecast_columns))

    forecast_pipeline = TimeSeriesForecastingPipeline(
        model=model,
        timestamp_column=timestamp_column,
        id_columns=id_columns,
        target_columns=forecast_columns,
    )

    context_length = model.config.context_length
    tsp = TimeSeriesPreprocessor.from_pretrained(prep_path)
    test_start_index = kwargs["test_start_index"] - context_length
    test_end_index = kwargs["test_end_index"]
    data = csv_to_df(kwargs)

    test_data = select_by_index(
        data,
        id_columns=id_columns,
        start_index=test_start_index,
        end_index=test_end_index,
    )
    test_data = tsp.preprocess(test_data)
    # tsp = TimeSeriesPreprocessor.from_pretrained(kwargs["preprocessor_path"])
    forecasts = forecast_pipeline(test_data)
    return test_data, forecasts


def create_figure(**kwargs) -> graph_objs.Figure:
    """Create a figure with with parameters given in kwargs (we do this for maximum implemntation flexibilty).

    At a minimum kwargs should contain the following:

    Returns:
        graph_objs.Figure: The figure that will get displayed in the UI
    """
    # finetuned_model_path = kwargs["finetuned_model_path"]
    forecast_columns = kwargs["forecast_columns"]
    model_path = get_model_path(**kwargs)
    # print("MODEL PATH", model_path)
    timestamp_column = kwargs["timestamp_column"]

    model_class = get_model_class(model_path)

    model = model_class.from_pretrained(model_path, num_input_channels=len(forecast_columns))
    context_length = model.config.context_length
    periodicity = kwargs["periodicity"]
    channel = kwargs["channel"]

    # we want to remove some parameters that
    # do not get used within our forecast method
    # because passing them will break the cache from
    # call to call
    cache_breakers = ["channel"]
    cachable_args = copy.deepcopy(kwargs)
    _ = [cachable_args.pop(arg) for arg in cache_breakers]

    test_data, forecasts = forecast(**cachable_args)

    answer = plot_ts_forecasting(
        test_data,
        forecasts,
        forecast_columns=[channel],
        timestamp_column=timestamp_column,
        periodicity=periodicity,
        prediction_length=model.config.prediction_length,
        context_length=context_length,
        plot_start=0,  # -4 * 24 * 2 - 1,
        plot_end=context_length + model.config.prediction_length * 3,
        plot_stride=model.config.prediction_length,
        num_predictions=3,
        title=channel,
        fig_size=(1600, 200),
        plot_type="plotly",
        return_image=False,
    )

    answer.update_layout(autosize=False, width=500, height=500)
    return answer



# Copyright contributors to the TSFM project
#
import bisect
from typing import Union

import numpy as np
import torch
from gluonts.dataset.split import InputDataset, LabelDataset, TrainingDataset
from gluonts.itertools import batcher
from gluonts.transform.feature import LastValueImputation
from torch.utils.data import Dataset
from tqdm import tqdm

from tsfm_public.toolkit.dataset import _torch
from tsfm_public.toolkit.time_series_preprocessor import DEFAULT_FREQUENCY_MAPPING


# TTM Constants:
TTM_MAX_FORECAST_HORIZON = 720
TTM_MIN_FORECAST_HORIZON = 16

RESOLUTION_MAP = {
    "oov": "oov",
    "OOV": "oov",
    "min": "min",
    "1min": "min",
    "T": "min",
    "1T": "min",
    "2min": "2min",
    "2T": "2min",
    "5min": "5min",
    "5T": "5min",
    "10min": "10min",
    "10T": "10min",
    "15min": "15min",
    "15T": "15min",
    "30min": "30min",
    "30T": "30min",
    "h": "h",
    "1h": "h",
    "H": "h",
    "1H": "h",
    "d": "d",
    "1d": "d",
    "D": "d",
    "1D": "d",
    "w": "W",
    "1w": "W",
    "W": "W",
    "1W": "W",
    "W-FRI": "W",
    "W-TUE": "W",
    "W-MON": "W",
    "W-WED": "W",
    "W-THU": "W",
    "W-SAT": "W",
    "W-SUN": "W",
    "M": "W",
    "1M": "W",
    "Q-DEC": "W",
    "A-DEC": "A",
}


def get_freq_mapping():
    freq_token_mapping = {}

    freq_token_mapping["oov"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["oov"]])
    freq_token_mapping["T"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["min"]])
    freq_token_mapping["2T"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["2min"]])
    freq_token_mapping["5T"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["5min"]])
    freq_token_mapping["10T"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["10min"]])
    freq_token_mapping["15T"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["15min"]])
    freq_token_mapping["30T"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["30min"]])
    freq_token_mapping["H"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["h"]])
    freq_token_mapping["D"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["d"]])
    freq_token_mapping["W"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["W"]])
    freq_token_mapping["W-FRI"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["W"]])
    freq_token_mapping["W-TUE"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["W"]])
    freq_token_mapping["W-MON"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["W"]])
    freq_token_mapping["W-WED"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["W"]])
    freq_token_mapping["W-THU"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["W"]])
    freq_token_mapping["W-SAT"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["W"]])
    freq_token_mapping["W-SUN"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["W"]])
    freq_token_mapping["M"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["W"]])
    freq_token_mapping["A-DEC"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["W"]])
    freq_token_mapping["Q-DEC"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["W"]])

    # Seconds are currently mapped to OOV
    freq_token_mapping["10S"] = torch.Tensor([DEFAULT_FREQUENCY_MAPPING["oov"]])

    return freq_token_mapping


def impute_series(target):
    if np.isnan(target).any():
        target = target.copy()
        if len(target.shape) == 2:
            for i in range(target.shape[0]):
                target[i, ...] = LastValueImputation()(target[i, ...])
        elif len(target.shape) == 1:
            target = LastValueImputation()(target)
        else:
            raise Exception("Only 1D and 2D arrays are accepted by the impute_series() function.")
    return target


class StandardScalingGluonTSDataset:
    """
    TTM works best on standard scaled data, especially if fewshot
    finetuning is being performed.
    We can utilize the entire available context to do that.
    This is a global sclaing operation done independently on
    each channel.
    """

    def __init__(self) -> None:
        self.mean = {}
        self.std = {}

    def fit(self, train_data: Union[TrainingDataset, InputDataset]):
        """Calculate the statistics on the historical train data.

        Args:
            train_data (Union[TrainingDataset, InputDataset]): Iterator with
                each series of shape [num_channels, context_length] for multivariate
                and [context_length] for univariate.
        """
        for batch in tqdm(batcher(train_data, batch_size=1)):
            if batch[0]["target"].ndim == 1:
                batch[0]["target"] = batch[0]["target"].reshape(1, -1)  # [1, context_length]
            self.mean[batch[0]["item_id"]] = np.mean(impute_series(batch[0]["target"]), axis=1).reshape(-1, 1)
            std = np.std(impute_series(batch[0]["target"]), axis=1).reshape(-1, 1)
            for i in range(std.shape[0]):
                if std[i] == 0:
                    std[i] = 1
            self.std[batch[0]["item_id"]] = std

    def transform(self, data: Union[TrainingDataset, InputDataset]):
        """Apply scaler using calculated statistics.

        Args:
            data (Union[TrainingDataset, InputDataset]): Iterator with
                each series of shape [num_channels, context_length] for multivariate
                and [context_length] for univariate.

        Returns:
            Iternator: With each series transformed.
        """
        assert len(self.mean) > 0
        assert len(self.std) > 0
        out = list(data)
        for i, _ in tqdm(enumerate(out)):
            item_id = out[i]["item_id"]
            out[i]["target"] = (impute_series(out[i]["target"]) - self.mean[item_id]) / (self.std[item_id])
        return iter(out)

    def inverse_transform(
        self, data: np.ndarray, series_ids: list, prediction_channel_indices: list = []
    ) -> np.ndarray:
        """Inverse transform, and bring data to original scale.

        Args:
            data (np.ndarray): Forecast output of shape [batch, context_length, num_channels]
            series_ids (list): List of item ids.
            prediction_channel_indices (list, optional): List of target channel indices. Defaults to [].

        Raises:
            Exception: If NaN is found in the forecast.

        Returns:
            np.ndarray: Of shape [batch, context_length, num_channels].
        """
        out = np.zeros(data.shape)

        for i in tqdm(range((data.shape[0]))):
            if len(prediction_channel_indices) > 0:
                out[i, ...] = (
                    data[i, ...] * (self.std[series_ids[i]][prediction_channel_indices].T)
                    + self.mean[series_ids[i]][prediction_channel_indices].T
                )
            else:
                out[i, ...] = data[i, ...] * (self.std[series_ids[i]].T) + self.mean[series_ids[i]].T
            if np.isnan(out[i, ...]).any():
                raise Exception("NaN found in forecast!")
        return out


class TorchDatasetFromGluonTSTrainingDataset(Dataset):
    def __init__(
        self,
        gluon_dataset: TrainingDataset,
        context_length: int,
        prediction_length: int,
        last_window_only=False,
        gen_more_samples_for_short_series: bool = True,
        force_short_context: bool = False,
        min_context_mult: int = 4,
        fewshot_fraction: float = 1.0,
        fewshot_location: str = "rand",  # end/start
        use_mask: bool = False,
        send_freq: bool = True,
        freq: str = None,
    ):
        """Wrapper to create pytorch `Dataset` from GluonTS dataset.

        Args:
            gluon_dataset (TrainingDataset): Training dataset.
            context_length (int): Context length.
            prediction_length (int): Prediction length.
            last_window_only (bool, optional): If True, only last window will be take from each series.
                Defaults to False.
            gen_more_samples_for_short_series (bool, optional): If True, it will pad zeros and try to
                generate more samples for short series. Defaults to True.
            force_short_context (bool, optional): If True, it will force to mine short context.
                Defaults to False (recommended).
            min_context_mult (int, optional): Minimum context length multiplier (of prediction length)
                to encourage more sample generation for short series. Defaults to 4.
            fewshot_fraction (float, optional): Fewshot fraction. Defaults to 1.0.
            fewshot_location (str, optional): Fewshot location. Defaults to "rand" (recommended).
            send_freq (bool, optional): Send frequency in getitem. Defaults to True.
            freq (str, optional): Frequency of the dataset. Defaults to None.
        """

        # assert context_length > prediction_length, f'sequence lenght {context_length} has to be strictly greater than forecast length {prediction_length}'
        self.context_length = context_length
        self.prediction_length = min(prediction_length, TTM_MAX_FORECAST_HORIZON)
        self.X = list(gluon_dataset)
        self.last_window_only = last_window_only
        self.stride = 1  # TODO: support other strides
        min_context_needed_mult = min_context_mult  # 4*H needed to forecast
        self.force_short_context = force_short_context
        self.use_mask = use_mask
        self.send_freq = send_freq
        self.freq = freq
        self.series_ids = []

        # force short context
        if self.force_short_context:
            self.actual_seq_len = context_length
            self.context_length = min_context_needed_mult * self.prediction_length
            gen_more_samples_for_short_series = False

        # handle univariate series, and nans
        for i, _ in enumerate(self.X):
            if self.X[i]["target"].ndim == 1:
                self.X[i]["target"] = self.X[i]["target"].reshape(1, -1)

            # Nan imputation
            self.X[i]["target"] = impute_series(self.X[i]["target"])

            # Fewshot: for fewshot_location `start` or `end` truncate each series
            if fewshot_fraction < 1.0 and fewshot_location in ["start", "end"]:
                len_ = self.X[i]["target"].shape[1]
                fewshot_len_ = int(np.floor(len_ * fewshot_fraction))
                if fewshot_len_ >= self.prediction_length * min_context_needed_mult:
                    if fewshot_location == "end":
                        self.X[i]["target"] = self.X[i]["target"][:, -fewshot_len_:]
                    elif fewshot_location == "start":
                        self.X[i]["target"] = self.X[i]["target"][:, :fewshot_len_]

            if self.X[i]["target"].shape[1] < self.context_length + self.prediction_length:
                # This means only 1 sample can be created from this series
                # even after zero-padding. We try to create more when
                # `gen_more_samples_for_short_series=True`
                if (
                    gen_more_samples_for_short_series
                    and self.X[i]["target"].shape[1] >= (min_context_needed_mult + 1) * self.prediction_length
                ):
                    # make sure at least a context of min_context_needed_mult*H is possible
                    # pad more zeros to create more training samples
                    pad = np.zeros(
                        (
                            self.X[i]["target"].shape[0],
                            self.context_length - min_context_needed_mult * self.prediction_length,
                        )
                    )
                else:
                    # pad it to Seq_len + Forecast_len to create 1 window
                    pad = np.zeros(
                        (
                            self.X[i]["target"].shape[0],
                            self.context_length + self.prediction_length - self.X[i]["target"].shape[1],
                        )
                    )

                # prepend
                self.X[i]["target"] = np.concatenate((pad, self.X[i]["target"]), axis=1)
                # print(self.X[i]["target"].shape)

            # series id
            self.series_ids.append(self.X[i]["item_id"])

        # get shape
        if not self.last_window_only:
            self.cumulative_sizes = self.cumsum(self.X)

    def cumsum(self, list_data):
        """
        list_data: list of numpy array of shape [channels x len]
        """
        list_len, sum_ = [], 0
        for i, elm in enumerate(list_data):
            data = elm["target"]
            len_ = data.shape[1] - self.context_length - self.prediction_length + 1
            list_len.append(len_ + sum_)
            sum_ += len_
        return list_len

    def __len__(self):
        if self.last_window_only:
            return len(self.X)  # = num of series
        else:
            return self.cumulative_sizes[-1] // self.stride

    def __getitem__(self, idx):
        if self.last_window_only:
            seq_x = self.X[idx]["target"][:, -(self.context_length + self.prediction_length) : -self.prediction_length]
            seq_y = self.X[idx]["target"][:, -(self.prediction_length) :]
            item_id = self.series_ids[idx]
        else:
            idx = idx * self.stride
            if idx < 0:
                if -idx > len(self):
                    raise ValueError("absolute value of index should not exceed dataset length")
                idx = len(self) + idx
            series_idx = bisect.bisect_right(self.cumulative_sizes, idx)
            if series_idx == 0:
                time_id = idx
            else:
                time_id = idx - self.cumulative_sizes[series_idx - 1]
            seq_x = self.X[series_idx]["target"][:, time_id : time_id + self.context_length]
            seq_y = self.X[series_idx]["target"][
                :, time_id + self.context_length : time_id + self.context_length + self.prediction_length
            ]
            item_id = self.series_ids[series_idx]

        if self.force_short_context and seq_x.shape[1] < self.actual_seq_len:
            pad = np.zeros((seq_x.shape[0], self.actual_seq_len - seq_x.shape[1]))
            seq_x = np.concatenate((pad, seq_x), axis=1)
            past_observed_mask = np.ones(seq_x.shape).astype(bool)
            past_observed_mask[:, : pad.shape[1]] = False
        else:
            # Create a boolean mask where non-zero values are True
            nonzero_mask = seq_x != 0
            # Find the first non-zero column index
            col_indices = np.where(nonzero_mask.any(axis=0))[0]
            if len(col_indices) == 0:
                past_observed_mask = np.zeros(seq_x.shape).astype(bool)
            else:
                past_observed_mask = np.ones(seq_x.shape).astype(bool)
                past_observed_mask[:, : col_indices[0]] = False

        # return torch.from_numpy(seq_x.astype(np.float)).float()
        seq_x, seq_y = _torch(seq_x, seq_y)

        return_output = {
            "past_values": seq_x.T,
            "future_values": seq_y.T,
            "item_id": item_id,
        }
        if self.use_mask:
            return_output["past_observed_mask"]: past_observed_mask.T

        if self.send_freq:
            freq_map = get_freq_mapping()
            return_output["freq_token"] = freq_map[self.freq]

        return return_output


class TorchDatasetFromGluonTSTestDataset(Dataset):
    def __init__(
        self,
        gluon_test_input: InputDataset,
        gluon_test_label: LabelDataset,
        context_length: int,
        prediction_length: int,
        force_short_context: bool = False,
        min_context_mult: int = 4,
    ):
        """Wrapper to create pytorch `Dataset` from GluonTS dataset.

        Args:
            gluon_test_input (InputDataset): GluonTS input dataset.
            gluon_test_label (LabelDataset): GluonTS label dataset.
            context_length (int): Context length.
            prediction_length (int): Forecast horizon.
            force_short_context (bool, optional): If True, it will force to mine short context.
                Defaults to False (recommended).
            min_context_mult (int, optional): Minimum context length multiplier (of prediction length)
                to encourage more sample generation for short series. Defaults to 4.
        """
        # assert context_length > prediction_length, f'sequence lenght {context_length} has to be strictly greater than forecast length {prediction_length}'
        self.context_length = context_length
        self.prediction_length = prediction_length
        self.X = list(gluon_test_input)
        self.Y = list(gluon_test_label)
        self.min_context_needed_mult = min_context_mult
        self.force_short_context = force_short_context
        if self.force_short_context:
            self.actual_seq_len = context_length
            self.context_length = self.min_context_needed_mult * self.prediction_length

    def __len__(self):
        return len(self.Y)

    def __getitem__(self, idx):
        seq_x = self.X[idx]["target"]
        seq_y = self.Y[idx]["target"]

        if len(seq_x.shape) == 1:
            seq_x = seq_x.reshape(1, -1)
            seq_y = seq_y.reshape(1, -1)

        if seq_x.shape[1] < self.context_length:
            pad = np.zeros((seq_x.shape[0], self.context_length - seq_x.shape[1]))
            seq_x = np.concatenate((pad, seq_x), axis=1)

        seq_x, seq_y = _torch(seq_x[:, -self.context_length :], seq_y[:, : self.prediction_length])

        if self.force_short_context and seq_x.shape[1] < self.actual_seq_len:
            pad = np.zeros((seq_x.shape[0], self.actual_seq_len - seq_x.shape[1]))
            seq_x = np.concatenate((pad, seq_x), axis=1)
            past_observed_mask = np.ones(seq_x.shape).astype(bool)
            past_observed_mask[:, : pad.shape[1]] = False
        else:
            # Create a boolean mask where non-zero values are True
            nonzero_mask = seq_x != 0
            # Find the first non-zero column index
            col_indices = np.where(nonzero_mask.any(axis=0))[0]
            if len(col_indices) == 0:
                past_observed_mask = np.zeros(seq_x.shape).astype(bool)
            else:
                past_observed_mask = np.ones(seq_x.shape).astype(bool)
                past_observed_mask[:, : col_indices[0]] = False

        return_output = {
            "past_values": seq_x.T,
            "future_values": seq_y.T,
        }
        if self.use_mask:
            return_output["past_observed_mask"]: past_observed_mask.T

        return return_output


class ForecastDataset(Dataset):
    def __init__(self, forecast_samples, series_ids, insample_errors, point_forecasts, quantiles):
        self.forecast_samples = forecast_samples
        self.series_ids = series_ids
        self.insample_errors = insample_errors
        self.point_forecasts = point_forecasts
        self.quantiles = quantiles

    def __len__(self):
        return self.forecast_samples.shape[0]

    def __getitem__(self, idx):
        forecast_sample = self.forecast_samples[idx]
        series_id = self.series_ids[idx]
        insample_error = self.insample_errors[series_id]
        point_forecast = self.point_forecasts[idx]
        return forecast_sample, insample_error, point_forecast



import logging
import os


def write_termination_log(text, log_file="error.log"):
    """Writes text to termination log.

    Args:
        text: str
        log_file: Optional[str]
    """
    log_file = os.environ.get("TERMINATION_LOG_FILE", log_file)
    try:
        with open(log_file, "a", encoding="utf-8") as handle:
            handle.write(text)
    except Exception as e:  # pylint: disable=broad-except
        logging.warning(f"Unable to write termination log due to {e}")



# Copyright contributors to the TSFM project
#

import json
import os
from typing import Any, Dict

import pandas as pd
import requests


def get_tuple(resp):
    resp = resp.json()
    df = [pd.DataFrame.from_dict(r) for r in resp["results"]]
    return df, {k: v for k, v in resp.items() if "data_point" in k}


def get_inference_response(
    msg: Dict[str, Any],
    url: str = "http://127.0.0.1:8000/v1/inference/forecasting",
) -> requests.Response:
    headers = {}
    return requests.post(url, json=msg, headers=headers)


def numts(msg):
    return 1 if "id_cols" not in msg["parameters"] else len(set(msg["parameters"]["id_columns"]))


"""
Scenario Outline: should get a 200 response code and 10 rows from request for model "granite-ttm-512-96-r2" with 512 time points input # features/time_series_forecast.feature:123
    And the response result length should be 10 # features/time_series_forecast.feature:137
      Error: Actual HUFL length is not equal to 10.
"""


def test_scenario_01():
    """Comments:
    The payload as given to me from product asks for only two predictions,
    not 10, as shown in the
    """
    msg = json.load(open(os.path.join(os.path.dirname(__file__), "prodpayload.json")))
    # change prediction length
    msg["parameters"]["prediction_length"] = 10
    resp = get_inference_response(msg)

    assert resp.ok
    df, _ = get_tuple(resp)
    numtseries = numts(msg)
    for idx in range(numtseries):
        assert df[idx].shape[0] == msg["parameters"]["prediction_length"]
        assert df[idx]["HUFL"].shape[0] == msg["parameters"]["prediction_length"]


"""
  Scenario Outline: should get a 400 response code from request for model "granite-ttm-512-96-r2" with 512 time points input and 1024 prediction_length # features/time_series_forecast.feature:145
    Then the response code should be 400 # features/time_series_forecast.feature:158
      Error: expected response code to be: 400, but actual is: 200.
"""


def test_scenario_02():
    msg = json.load(open(os.path.join(os.path.dirname(__file__), "prodpayload.json")))
    # change prediction length
    msg["parameters"]["prediction_length"] = 1024
    resp: requests.Response = get_inference_response(msg)
    assert resp.status_code == 400



# Copyright contributors to the TSFM project
#
"""Hugging Face Pipeline for Time Series Tasks"""

from typing import Any, Dict, List, Union

import pandas as pd
from transformers import PreTrainedModel
from transformers.pipelines.base import (
    GenericTensor,
    build_pipeline_init_args,
)
from transformers.utils import add_end_docstrings, logging

from .dataset import ClassificationDFDataset
from .time_series_forecasting_pipeline import TimeSeriesPipeline


# Eventually we should support all time series models
# MODEL_FOR_TIME_SERIES_FORECASTING_MAPPING_NAMES = OrderedDict(
#     [
#         # Model for Time Series Forecasting
#         ("PatchTST", "PatchTSTForPrediction"),
#         ("TST", "TimeSeriesTransformerForPrediction"),
#     ]
# )


logger = logging.get_logger(__name__)


@add_end_docstrings(
    build_pipeline_init_args(has_tokenizer=False, has_feature_extractor=True, has_image_processor=False)
)
class TimeSeriesClassificationPipeline(TimeSeriesPipeline):
    """
    Time Series Classification using HF time series forecasting models. This pipeline consumes a `pandas.DataFrame`
    containing the time series data and produces a new `pandas.DataFrame` containing the classification results.

    """

    def __init__(
        self,
        model: Union["PreTrainedModel"],
        *args,
        explode_forecasts: bool = False,
        inverse_scale_outputs: bool = True,
        add_known_ground_truth: bool = True,
        **kwargs,
    ):
        # autopopulate from feature extractor and model
        if "feature_extractor" in kwargs:
            for p in [
                "id_columns",
                "timestamp_column",
                "input_columns",
                "label_column",
                # "observable_columns",
                # "control_columns",
                # "conditional_columns",
                # "categorical_columns",
                "static_categorical_columns",
                # "freq",
            ]:
                if p not in kwargs:
                    kwargs[p] = getattr(kwargs["feature_extractor"], p)

        if "context_length" not in kwargs:
            kwargs["context_length"] = model.config.context_length

        # check if we need to use the frequency token, get token if needed
        # use_frequency_token = getattr(model.config, "resolution_prefix_tuning", False)

        # if use_frequency_token and "feature_extractor" in kwargs:
        #     kwargs["frequency_token"] = kwargs["feature_extractor"].get_frequency_token(kwargs["freq"])
        # else:
        #     kwargs["frequency_token"] = None

        super().__init__(model, *args, **kwargs)

        self.__context_memory = {}

        if self.framework == "tf":
            raise ValueError(f"The {self.__class__} is only available in PyTorch.")

        # self.check_model_type(MODEL_FOR_TIME_SERIES_FORECASTING_MAPPING)

    def _sanitize_parameters(
        self,
        **kwargs,
    ):
        """Assigns parameters to the different steps of the process. If context_length and prediction_length
        are not provided they are taken from the model config.

        For expected parameters see the call method below.
        """

        preprocess_kwargs = {}
        postprocess_kwargs = {}

        preprocess_params = [
            "prediction_length",
            "context_length",
            # "frequency_token",
            "id_columns",
            "timestamp_column",
            "input_columns",
            "label_column",
            "static_categorical_columns",
        ]
        postprocess_params = [
            "prediction_length",
            "context_length",
            "id_columns",
            "timestamp_column",
            "input_columns",
            "label_column",
            "static_categorical_columns",
            # "freq",
        ]

        for c in preprocess_params:
            if c in kwargs:
                preprocess_kwargs[c] = kwargs[c]

        for c in postprocess_params:
            if c in kwargs:
                postprocess_kwargs[c] = kwargs[c]

        # same logic as HF Pipeline
        batch_size = kwargs.get("batch_size", self._batch_size)
        if batch_size is None:
            if self._batch_size is None:
                batch_size = 1
            else:
                batch_size = self._batch_size

        num_workers = kwargs.get("num_workers", self._num_workers)
        if num_workers is None:
            if self._num_workers is None:
                num_workers = 0
            else:
                num_workers = self._num_workers

        forward_kwargs = {"batch_size": batch_size, "num_workers": num_workers}

        return preprocess_kwargs, forward_kwargs, postprocess_kwargs

    def __call__(
        self,
        time_series: Union["pd.DataFrame", str],
        **kwargs,
    ):
        """Main method of the forecasting pipeline. Takes the input time series data (in tabular format) and
        produces predictions.

        Args:
            time_series (Union[&quot;pandas.DataFrame&quot;, str]): A pandas dataframe containing the time series on
                which to perform inference.

        Keyword arguments:
            future_time_series (Union[&quot;pandas.DataFrame&quot;, str]): A pandas dataframe containing future values,
                i.e., exogenous or supporting features which are known in advance.

            feature_extractor (TimeSeriesPreprocessor): A time series preprpocessor object that specifies how the time
                series should be prepared. If this is provided, any of the other options below will be automatically
                populated from this instance.

            timestamp_column (str): The name of the column containing the timestamp of the time series.

            id_columns (List[str]): List of column names which identify different time series in a multi-time series input.

            input_columns (List[str]): List of column names which identify the target channels in the input, these are the
                columns that will be forecasted.

            label_column (str): Column containing time series label

            static_categorical_columns (List[str]): List of column names which identify categorical-valued channels in the input
                which are fixed over time.

            freq (str): A freqency indicator for the given `timestamp_column`. See
                https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#period-aliases for a description of the
                allowed values. If not provided, we will attempt to infer it from the data. If not provided, frequency will be
                inferred from `timestamp_column`.

            prediction_length (int): The length of the desired forecast. Currently, this value must not exceed the maximum value
                suported by the model. If not specified, the maximum value supported by the model is used.

            context_length (int): Specifies the length of the context windows extracted from the historical data for feeding into
                the model.

            explode_forecasts (bool): If true, forecasts are returned one value per row of the pandas dataframe. If false, the
                forecast over the prediction length will be contained as a list in a single row of the pandas dataframe.

            inverse_scale_outputs (bool): If true and a valid feature extractor is provided, the outputs will be inverse scaled.

            add_known_ground_truth (bool): If True add columns containing the ground truth data. Prediction columns will have a
                suffix of "_prediction". Default True. If false, on columns containing predictions are produced, no suffix is
                added.

        Return (pandas dataframe):
            A new pandas dataframe containing the forecasts. Each row will contain the id, timestamp, the original
            input feature values and the output forecast for each input column. The output forecast is a list containing
            all the values over the prediction horizon.

        """

        return super().__call__(time_series, **kwargs)

    def preprocess(self, time_series, **kwargs) -> Dict[str, Union[GenericTensor, List[Any]]]:
        """Preprocess step
        Load the data, if not already loaded, and then generate a pytorch dataset.
        """

        timestamp_column = kwargs.get("timestamp_column")
        # id_columns = kwargs.get("id_columns")
        # input_columns = kwargs.get("input_columns")
        # label_column = kwargs.get("label_column")
        # context_length = kwargs.get("context_length")

        # use the feature extractor here

        if isinstance(time_series, str):
            time_series = pd.read_csv(
                time_series,
                parse_dates=[timestamp_column],
            )

        if self.feature_extractor:
            time_series_prep = self.feature_extractor.preprocess(time_series)
        else:
            time_series_prep = time_series

        # add full_series option here
        kwargs["full_series"] = True

        # prepare classification dataset
        dataset = ClassificationDFDataset(
            time_series_prep,
            **kwargs,
        )

        self.__context_memory["data"] = time_series

        return {"dataset": dataset}

    def _forward(self, model_inputs, **kwargs):
        """Forward step
        Responsible for taking pre-processed dictionary of tensors and passing it to
        the model. Aligns model parameters with the proper input parameters. Only passes
        the needed parameters from the dictionary to the model, but adds them back to the
        ouput for the next step.

        The keys in model_outputs are governed by the underlying model combined with any
        original input keys.
        """

        model_outputs = self.model(**model_inputs)

        return model_outputs

    def postprocess(self, input, **kwargs):
        """Postprocess step
        Takes the dictionary of outputs from the previous step and converts to a more user
        readable pandas format.

        If the explode forecasts option is True, then individual forecasts are expanded as multiple
        rows in the dataframe. This should only be used when producing a single forecast (i.e., unexploded
        result is one row per ID).
        """
        out = {}

        model_output_key = "prediction_outputs"  #  if "prediction_outputs" in input.keys() else "prediction_logits"

        # name the predictions of target columns
        # outputs should only have size equal to target columns

        if kwargs.get("timestamp_column", None):
            out[kwargs["timestamp_column"]] = input["timestamp"]
        for i, c in enumerate(kwargs["id_columns"]):
            out[c] = [elem[i] for elem in input["id"]]

        # input series
        for c in kwargs["input_columns"]:
            input_data = self.__context_memory["data"]
            out[c] = input_data[c]

        # ground truth
        column = kwargs["label_column"]
        out[column] = input["target_values"].numpy().tolist()

        # predictions
        column = f"{kwargs['label_column']}_prediction"
        out[column] = input[model_output_key].numpy().argmax(axis=1).tolist()

        out = pd.DataFrame(out)

        # inverse transform labels
        if self.feature_extractor is not None:
            out = self.feature_extractor.inverse_transform_labels(out)
            out = self.feature_extractor.inverse_transform_labels(out, suffix="_prediction")

        self.__context_memory = {}
        return out



# Copyright contributors to the TSFM project
#
"""Preprocessor for time series data preparation"""

import copy
import datetime
import enum
import json
import logging
from collections import Counter
from typing import Any, Dict, Generator, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from datasets import Dataset
from pandas.tseries.frequencies import to_offset
from sklearn.preprocessing import LabelEncoder as LabelEncoder_
from sklearn.preprocessing import MinMaxScaler as MinMaxScaler_
from sklearn.preprocessing import OrdinalEncoder as OrdinalEncoder_
from sklearn.preprocessing import StandardScaler as StandardScaler_
from torch.utils.data import Subset
from transformers.feature_extraction_utils import (
    PreTrainedFeatureExtractor,
)

from .dataset import ForecastDFDataset
from .processor import BaseProcessor
from .util import (
    FractionLocation,
    convert_to_univariate,
    get_split_params,
    join_list_without_repeat,
    select_by_fixed_fraction,
)


LOGGER = logging.getLogger(__file__)


INTERNAL_ID_COLUMN = "__id"
INTERNAL_ID_VALUE = "0"

DEFAULT_FREQUENCY_MAPPING = {
    "oov": 0,
    "min": 1,  # minutely
    "2min": 2,
    "5min": 3,
    "10min": 4,
    "15min": 5,
    "30min": 6,
    "h": 7,  # hourly
    "H": 7,  # hourly, for compatibility
    "d": 8,  # daily, for compatibility
    "D": 8,  # daily
    "W": 9,  # weekly
}

TYPE_TO_STRING = {int: "int", np.int64: "numpy.int64", str: "str", float: "float", np.float64: "numpy.float64"}
STRING_TO_TYPE = {_v: _k for _k, _v in TYPE_TO_STRING.items()}


class SKLearnFeatureExtractionBase:
    """Simple wrapper class to adapt Sklearn functions to work with the HF
    serialization approach.
    """

    def to_dict(self) -> Dict[str, Any]:
        """Return a dictionary of parameters from which we can reconstruct"""
        return self.__getstate__()

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_dict(cls, feature_extractor_dict: Dict[str, Any], **kwargs) -> "SKLearnFeatureExtractionBase":
        """ """

        t = cls()
        t.__setstate__(feature_extractor_dict)

        return t


class StandardScaler(StandardScaler_, SKLearnFeatureExtractionBase):
    """Simple wrapper class to adapt standard scaler to work with the HF
    serialization approach.
    """


class MinMaxScaler(MinMaxScaler_, SKLearnFeatureExtractionBase):
    """Simple wrapper class to adapt min/max scaler to work with the HF
    serialization approach.
    """


class OrdinalEncoder(OrdinalEncoder_, SKLearnFeatureExtractionBase):
    """Simple wrapper class to adapt OrdinalEncoder to work with the HF
    serialization approach.
    """


class LabelEncoder(LabelEncoder_, SKLearnFeatureExtractionBase):
    """Simple wrapper class to adapt LabelEncoder to work with the HF
    serialization approach.
    """


class TimeSeriesTask(enum.Enum):
    """`Enum` for the different kinds of time series datasets we need to create."""

    CLASSIFICATION = "classification"
    MASKED_PRETRAINING = "mask_pretraining"
    FORECASTING = "forecasting"
    REGRESSION = "regression"


class ScalerType(enum.Enum):
    """`Enum` for the different kinds of scalers."""

    MINMAX = "minmax"
    STANDARD = "standard"


class TimeSeriesProcessorBase(BaseProcessor):
    PROCESSOR_NAME = "preprocessor_config.json"

    # HF serialization support methods
    def to_dict(self) -> Dict[str, Any]:
        """
        Serializes this instance to a Python dictionary.

        Returns:
            `Dict[str, Any]`: Dictionary of all the attributes that make up this feature extractor instance.
        """
        output = super().to_dict()

        for k, v in output["scaler_dict"].items():
            output["scaler_dict"][k] = v.to_dict()

        if "target_scaler_dict" in output:
            for k, v in output["target_scaler_dict"].items():
                output["target_scaler_dict"][k] = v.to_dict()
            input_scaler_dict = self.target_scaler_dict
        else:
            input_scaler_dict = self.scaler_dict

        # get type information; assume homogeneous types
        if self.scaling_id_columns and self.scaling:
            akey = next(iter(input_scaler_dict.keys()))
            if isinstance(akey, Tuple):
                key_types = [type(k) for k in akey]
            else:
                key_types = [type(akey)]
        else:
            key_types = []

        output["scaling_id_columns_types"] = [TYPE_TO_STRING[k] for k in key_types]

        if self.categorical_encoder:
            output["categorical_encoder"] = output["categorical_encoder"].to_dict()

        if "label_encoder" in output:
            output["label_encoder"] = output["label_encoder"].to_dict()

        return output

    def to_json_string(self) -> str:
        """
        Serializes this instance to a JSON string.

        Returns:
            `str`: String containing all the attributes that make up this feature_extractor instance in JSON format.
        """
        dictionary = self.to_dict()

        def serialize_np_scalar(value):
            if isinstance(value, np.integer):
                return int(value)
            if isinstance(value, np.floating):
                return float(value)
            return value

        def recursive_check_ndarray(dictionary):
            key_map = {}
            for key, value in dictionary.items():
                if isinstance(key, tuple):
                    new_key = json.dumps([serialize_np_scalar(k) for k in key])
                    key_map[key] = new_key

                if key == "dtype":
                    # to do: ensure deserializable
                    dictionary[key] = value.__name__
                elif isinstance(value, np.ndarray):
                    dictionary[key] = value.tolist()
                elif isinstance(value, (np.integer, np.floating)):
                    dictionary[key] = serialize_np_scalar(value)
                elif isinstance(value, list):
                    dictionary[key] = [vv.tolist() if isinstance(vv, np.ndarray) else vv for vv in value]
                elif isinstance(value, dict):
                    dictionary[key] = recursive_check_ndarray(value)

            for key, new_key in key_map.items():
                dictionary[new_key] = dictionary.pop(key)

            return dictionary

        dictionary = recursive_check_ndarray(dictionary)

        # make sure private name "_processor_class" is correctly
        # saved as "processor_class"
        _processor_class = dictionary.pop("_processor_class", None)
        if _processor_class is not None:
            dictionary["processor_class"] = _processor_class

        return json.dumps(dictionary, indent=2, sort_keys=True) + "\n"

    @classmethod
    def from_dict(cls, feature_extractor_dict: Dict[str, Any], **kwargs) -> "PreTrainedFeatureExtractor":
        """
        Instantiates a type of [`~feature_extraction_utils.FeatureExtractionMixin`] from a Python dictionary of
        parameters.

        Args:
            feature_extractor_dict (`Dict[str, Any]`):
                Dictionary that will be used to instantiate the feature extractor object. Such a dictionary can be
                retrieved from a pretrained checkpoint by leveraging the
                [`~feature_extraction_utils.FeatureExtractionMixin.to_dict`] method.
            kwargs (`Dict[str, Any]`):
                Additional parameters from which to initialize the feature extractor object.

        Returns:
            [`~feature_extraction_utils.FeatureExtractionMixin`]: The feature extractor object instantiated from those
            parameters.
        """

        scaler_type = feature_extractor_dict.get("scaler_type", None)

        scaler_class = cls._get_scaler_class(scaler_type)
        id_types = feature_extractor_dict.get("scaling_id_columns_types", None)

        def deserialize_key(key, key_types=None):
            if not key_types:
                return key

            key = json.loads(key)
            if isinstance(key, (Tuple, List)):
                return tuple([STRING_TO_TYPE[k_type](k_item) for k_item, k_type in zip(key, key_types)])
            else:
                return STRING_TO_TYPE[key_types[0]](key)

        scaler_params = feature_extractor_dict.get("scaler_dict", None)
        if scaler_params is not None:
            scaler_params_copy = {}
            for k, v in scaler_params.items():
                scaler_params_copy[deserialize_key(k, key_types=id_types)] = scaler_class.from_dict(v)
            feature_extractor_dict["scaler_dict"] = scaler_params_copy

        target_scaler_params = feature_extractor_dict.get("target_scaler_dict", None)
        if target_scaler_params is not None:
            target_scaler_params_copy = {}
            for k, v in target_scaler_params.items():
                target_scaler_params_copy[deserialize_key(k, key_types=id_types)] = scaler_class.from_dict(v)
            feature_extractor_dict["target_scaler_dict"] = target_scaler_params_copy

        return super().from_dict(feature_extractor_dict, **kwargs)

    # support methods
    @classmethod
    def _get_scaler_class(cls, scaler_type):
        if scaler_type == ScalerType.MINMAX.value:
            return MinMaxScaler

        if scaler_type == ScalerType.STANDARD.value:
            return StandardScaler

        raise ValueError(f"Unknown scaler type {scaler_type} specified.")

    def _standardize_dataframe(
        self,
        dataset: Union[Dataset, pd.DataFrame],
    ) -> pd.DataFrame:
        """For given supported inputs, appropriately converts to a pandas dataframe. Adds an ID column
        if needed.

        Args:
            dataset (Union[Dataset, pd.DataFrame]): Input dataset

        Returns:
            pd.DataFrame: Converted dataframe with ID column.
        """
        if isinstance(dataset, Dataset):
            df = dataset.to_pandas()
        else:
            df = dataset.copy()

        # add id column when there are no id or scaling_id columns
        # or when scaling_id_columns == []
        if not self.id_columns or self.scaling_id_columns == []:
            df[INTERNAL_ID_COLUMN] = INTERNAL_ID_VALUE

        return df

    def _clean_up_dataframe(self, df: pd.DataFrame) -> None:
        """Removes columns added during internal processing of the provided dataframe.

        Currently, the following checks are done:
         - Remove INTERNAL_ID_COLUMN if present

        Args:
            df (pd.DataFrame): Input pandas dataframe

        Returns:
            pd.DataFrame: Cleaned up dataframe
        """

        if not self.id_columns:
            if INTERNAL_ID_COLUMN in df.columns:
                df.drop(columns=INTERNAL_ID_COLUMN, inplace=True)

    def _check_dataset(self, dataset: Union[Dataset, pd.DataFrame]):
        """Basic checks for input dataset.

        Args:
            dataset (Union[Dataset, pd.DataFrame]): Input time series data.

        Raises:
            ValueError: Raised if the dataset is empty.
        """
        if dataset is None or len(dataset) == 0:
            raise ValueError("Input dataset must not be null or zero length.")

        if self.id_columns:
            dtypes = dataset.dtypes
            for id in self.id_columns:
                if not (pd.api.types.is_string_dtype(dtypes[id]) or pd.api.types.is_integer_dtype(dtypes[id])):
                    raise ValueError(f"Data for identifier column {id} must be a string or integer type.")

    def _get_groups(
        self,
        dataset: pd.DataFrame,
    ) -> Generator[Tuple[Any, pd.DataFrame], None, None]:
        """Get groups of the time series dataset (multi-time series) based on the ID columns for scaling.
        Note that this is used for scaling purposes only.

        Args:
            dataset (pd.DataFrame): Input dataset

        Yields:
            Generator[Any, pd.DataFrame]: Group name and resulting pandas dataframe for the group.
        """
        if self.scaling_id_columns is not None and len(self.scaling_id_columns) > 0:
            group_by_columns = (
                self.scaling_id_columns if len(self.scaling_id_columns) > 1 else self.scaling_id_columns[0]
            )
        else:
            group_by_columns = INTERNAL_ID_COLUMN

        grps = dataset.groupby(by=group_by_columns)
        for name, g in grps:
            # g = g.sort_values(by=self.timestamp_column)
            yield name, g

    # training methods
    def _train_categorical_encoder(self, df: pd.DataFrame):
        cols_to_encode = self._get_columns_to_encode()

        if cols_to_encode:
            self.categorical_encoder = OrdinalEncoder()
            self.categorical_encoder.fit(df[cols_to_encode])

    # processing methods
    def _process_encoding(self, df: pd.DataFrame) -> pd.DataFrame:
        """Uses the trained categorical encoder to transform the categorical columns in the input dataframe.

        Args:
            df (pd.DataFrame): Input dataframe.

        Raises:
            RuntimeError: Raised when the encoder has not been trained.

        Returns:
            pd.DataFrame: The dataframe with the columns properly encoded.
        """
        cols_to_encode = self._get_columns_to_encode()
        if self.encode_categorical and cols_to_encode:
            if not self.categorical_encoder:
                raise RuntimeError("Attempt to encode categorical columns, but the encoder has not been trained yet.")
            df[cols_to_encode] = self.categorical_encoder.transform(df[cols_to_encode])
        return df


class TimeSeriesPreprocessor(TimeSeriesProcessorBase):
    """A preprocessor for supporting time series modeling tasks"""

    PROCESSOR_NAME = "preprocessor_config.json"

    def __init__(
        self,
        id_columns: List[str] = [],
        timestamp_column: Optional[str] = None,
        target_columns: List[str] = [],
        observable_columns: List[str] = [],
        control_columns: List[str] = [],
        conditional_columns: List[str] = [],
        categorical_columns: List[str] = [],
        static_categorical_columns: List[str] = [],
        context_length: int = 64,
        prediction_length: Optional[int] = None,
        scaling: bool = False,
        scaler_type: ScalerType = ScalerType.STANDARD.value,
        scaling_id_columns: Optional[List[str]] = None,
        encode_categorical: bool = True,
        time_series_task: str = TimeSeriesTask.FORECASTING.value,
        frequency_mapping: Dict[str, int] = DEFAULT_FREQUENCY_MAPPING,
        freq: Optional[Union[int, str]] = None,
        scale_categorical_columns: bool = True,
        **kwargs,
    ):
        """Multi-time series aware data preprocessor. Provides functions for scaling data and facilitates downstream
        operations on time series data, including model training and inference.

        Args:
            id_columns (List[str]): List of column names which identify different time series in a multi-time series input. Defaults to [].
            timestamp_column (Optional[str], optional): The name of the column containing the timestamp of the time series. Defaults to None.
            target_columns (List[str], optional): List of column names which identify the target channels in the input, these are the
                columns that will be forecasted. Defaults to [].
            observable_columns (List[str], optional): List of column names which identify the observable channels in the input.
                Observable channels are channels which we have knowledge about in the past and future. For example, weather
                conditions such as temperature or precipitation may be known or estimated in the future, but cannot be
                changed. Defaults to [].
            control_columns (List[str], optional): List of column names which identify the control channels in the input. Control
                channels are similar to observable channels, except that future values may be controlled. For example, discount
                percentage of a particular product is known and controllable in the future. Defaults to [].
            conditional_columns (List[str], optional): List of column names which identify the conditional channels in the input.
                Conditional channels are channels which we know in the past, but do not know in the future. Defaults to [].
            categorical_columns (List[str]): List of column names which identify time-varying categorical-valued channels in the input.
                Defaults to [].
            static_categorical_columns (List[str], optional): List of column names which identify categorical-valued channels in the input
                which are fixed over time. Defaults to [].
            context_length (int, optional): The length of the input context window. Defaults to 64.
            prediction_length (Optional[int], optional): The length of the prediction window. Defaults to None.
            scaling (bool, optional): If True, data is scaled. Defaults to False.
            scaler_type (ScalerType, optional): The type of scaling to perform. See ScalerType for available scalers. Defaults to ScalerType.STANDARD.value.
            scaling_id_columns (Optional[List[str]], optional): In some cases we need to separate data by a different set of id_columns
                when determining scaling factors. For the purposes of determining scaling, data will be grouped by the provided columns.
                If None, the `id_columns` will be used. If and empty list ([]), the dataset will be treated as a single group for scaling.
                Defaults to None. This should be a subset of the id_columns.
            encode_categorical (bool, optional): If True any categorical columns will be encoded using ordinal encoding. Defaults to True.
            time_series_task (str, optional): Reserved for future use. Defaults to TimeSeriesTask.FORECASTING.value.
            frequency_mapping (Dict[str, int], optional): A mapping which maps frequency strings to numerical values (integers). Defaults to DEFAULT_FREQUENCY_MAPPING.
            freq (Optional[Union[int, str]], optional): A frequency indicator for the given `timestamp_column`. See
                https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#period-aliases for a description of the
                allowed values. If not provided, we will attempt to infer it from the data. If not provided, frequency will be
                inferred from `timestamp_column`. Defaults to None.
            scale_categorical_columns (bool, optional): If True, the oridinal representations of categorical columns are scaled during preprocessing.
                Defaults to True.

        Raises:
            ValueError: Raised if `id_columns` is not a list.
            ValueError: Raised if `timestamp_column` is not a scalar.
        """
        # note base class __init__ method sets all arguments as attributes

        if not isinstance(id_columns, list):
            raise ValueError(f"Invalid argument provided for `id_columns`: {id_columns}")

        if isinstance(timestamp_column, list):
            raise ValueError(f"`timestamp_column` should not be a list, received: {timestamp_column}")

        self.id_columns = id_columns
        self.timestamp_column = timestamp_column
        self.target_columns = list(target_columns)
        self.observable_columns = list(observable_columns)
        self.control_columns = list(control_columns)
        self.conditional_columns = list(conditional_columns)
        self.categorical_columns = list(categorical_columns)
        self.static_categorical_columns = list(static_categorical_columns)

        self.context_length = context_length
        self.prediction_length = prediction_length
        self.scaling = scaling
        self.encode_categorical = encode_categorical
        self.time_series_task = time_series_task
        # self.scale_outputs = scale_outputs
        self.scaler_type = scaler_type

        # check subset
        if scaling_id_columns is not None:
            if not set(scaling_id_columns).issubset(self.id_columns):
                raise ValueError("`scaling_id_columns` must be a subset of `id_columns`")
            self.scaling_id_columns = scaling_id_columns
        else:
            self.scaling_id_columns = copy.copy(id_columns)

        # we maintain two scalers per time series to facilitate inverse scaling of the targets
        self.scaler_dict = {}
        self.target_scaler_dict = {}
        self.categorical_encoder = None
        self.frequency_mapping = frequency_mapping
        self.freq = freq
        self.scale_categorical_columns = scale_categorical_columns

        kwargs["processor_class"] = self.__class__.__name__

        super().__init__(**kwargs)

    def _validate_columns(self):
        """Check column specification parameters

        Raises:
            ValueError: Raised when a given column appears in multiple column specifiers.
        """

        dups = [
            k
            for k, v in Counter(
                self.target_columns
                + self.observable_columns
                + self.control_columns
                + self.conditional_columns
                + self.static_categorical_columns
            ).items()
            if v > 1
        ]
        if dups:
            raise ValueError(
                f"A column name should appear only once in `target_columns`, `observable_colums`, `control_columns`, `conditional_columns`, `categorical_columns`, and `static_categorical_columns`. Duplicates {dups}"
            )

        for c in self.categorical_columns:
            if all(
                c not in aset for aset in [self.conditional_columns, self.control_columns, self.observable_columns]
            ):
                raise ValueError(
                    "Each specified categorical column must also be included in one of 'conditional_columns', 'control_columns', or 'observable_columns'."
                )

    @classmethod
    def _get_scaler_class(cls, scaler_type):
        if scaler_type == ScalerType.MINMAX.value:
            return MinMaxScaler

        if scaler_type == ScalerType.STANDARD.value:
            return StandardScaler

        raise ValueError(f"Unknown scaler type {scaler_type} specified.")

    def _standardize_dataframe(
        self,
        dataset: Union[Dataset, pd.DataFrame],
    ) -> pd.DataFrame:
        """For given supported inputs, appropriately converts to a pandas dataframe. Adds an ID column
        if needed.

        Args:
            dataset (Union[Dataset, pd.DataFrame]): Input dataset

        Returns:
            pd.DataFrame: Converted dataframe with ID column.
        """
        if isinstance(dataset, Dataset):
            df = dataset.to_pandas()
        else:
            df = dataset.copy()

        # add id column when there are no id or scaling_id columns
        # or when scaling_id_columns == []
        if not self.id_columns or self.scaling_id_columns == []:
            df[INTERNAL_ID_COLUMN] = INTERNAL_ID_VALUE

        return df

    def _clean_up_dataframe(self, df: pd.DataFrame) -> None:
        """Removes columns added during internal processing of the provided dataframe.

        Currently, the following checks are done:
         - Remove INTERNAL_ID_COLUMN if present

        Args:
            df (pd.DataFrame): Input pandas dataframe

        Returns:
            pd.DataFrame: Cleaned up dataframe
        """

        if not self.id_columns:
            if INTERNAL_ID_COLUMN in df.columns:
                df.drop(columns=INTERNAL_ID_COLUMN, inplace=True)

    def _get_other_columns_to_scale(
        self,
    ) -> List[str]:
        """Returns the columns to perform scaling on, based on the options specified during
        preprocessor init.

        Returns:
            List[str]: List of column names
        """

        column_lists = [
            self.observable_columns,
            self.control_columns,
            self.conditional_columns,
        ]
        if self.scale_categorical_columns:
            column_lists.append(self.categorical_columns)
        cols_to_scale = join_list_without_repeat(*column_lists)

        return cols_to_scale

    def _get_columns_to_encode(
        self,
    ) -> List[str]:
        """Returns the columns to perform encoding on, based on the options specified during
        preprocessor init.

        Returns:
            List[str]: List of column names
        """
        cols_to_encode = self.static_categorical_columns + self.categorical_columns
        return cols_to_encode

    def _train_scaler(self, df: pd.DataFrame):
        cols_to_scale = self._get_other_columns_to_scale()
        scaler_class = self._get_scaler_class(self.scaler_type)

        if self.scaling:
            for name, g in self._get_groups(df):
                # check for na
                if (cols_to_scale and np.any(np.all(g[cols_to_scale].isna(), axis=0))) or np.any(
                    np.all(g[self.target_columns].isna(), axis=0)
                ):
                    raise RuntimeError(
                        "Input dataframe contains a group with at least one column containing all missing values."
                    )

                # train
                if cols_to_scale:
                    self.scaler_dict[name] = scaler_class()
                    self.scaler_dict[name].fit(g[cols_to_scale])

                self.target_scaler_dict[name] = scaler_class()
                self.target_scaler_dict[name].fit(g[self.target_columns])

    def get_frequency_token(self, token_name: str):
        token = self.frequency_mapping.get(token_name, None)
        if token is not None:
            return token

        # try to map as a frequency string
        try:
            token_name_offs = to_offset(token_name).freqstr
            token = self.frequency_mapping.get(token_name_offs, None)
            if token is not None:
                return token
        except ValueError:
            # lastly try to map the timedelta to a frequency string
            token_name_td = pd._libs.tslibs.timedeltas.Timedelta(token_name)
            token_name_offs = to_offset(token_name_td).freqstr
            token = self.frequency_mapping.get(token_name_offs, None)
            if token is not None:
                return token

        logging.warning(f"Frequency token {token_name} was not found in the frequency token mapping.")
        token = self.frequency_mapping["oov"]

        return token

    def _get_real_valued_dynamic_channels(
        self,
    ) -> List[str]:
        """Helper function to return list of the real-valued dynamic channels (columns)"""
        real_valued_dynamic_columns = join_list_without_repeat(
            self.target_columns,
            self.observable_columns,
            self.control_columns,
            self.conditional_columns,
        )
        return real_valued_dynamic_columns

    @property
    def num_input_channels(
        self,
    ) -> int:
        """Return the number of input channels

        Input channels are defined as those channels in:
            target_columns
            observable_columns
            control_columns
            conditional_columns

        Note that categorical columns should be specified as catetgorical_columns, and included in one of the above lists.
        """
        return len(self._get_real_valued_dynamic_channels())

    @property
    def exogenous_channel_indices(self) -> List[int]:
        """Return the indices of the exogenous columns

        In this case, exogenous are defined as control columns and observable columns. I.e., columns
        where we know the future values.
        """
        return [
            i
            for i, c in enumerate(self._get_real_valued_dynamic_channels())
            if c in self.control_columns + self.observable_columns
        ]

    @property
    def prediction_channel_indices(self) -> List[int]:
        """Return the indices of the prediction columns, i.e. targets"""
        return [i for i, c in enumerate(self._get_real_valued_dynamic_channels()) if c in self.target_columns]

    @property
    def categorical_vocab_size_list(self) -> List[int]:
        """Return the static_categorical_column vocabulary sizes."""
        if not self.static_categorical_columns or not self.encode_categorical:
            return None

        if not self.categorical_encoder:
            raise RuntimeError(
                "Vocabulary sizes are only available after training the preprocessor. Please run the `train` method first."
            )

        sizes = []
        for feat, cats in zip(self.categorical_encoder.feature_names_in_, self.categorical_encoder.categories_):
            if feat in self.static_categorical_columns:
                sizes.append(len(cats))

        return sizes

    def _set_targets(self, dataset: pd.DataFrame) -> None:
        if self.target_columns == []:
            skip_columns = copy.copy(self.id_columns) + [INTERNAL_ID_COLUMN]
            if self.timestamp_column:
                skip_columns.append(self.timestamp_column)

            skip_columns.extend(self.observable_columns)
            skip_columns.extend(self.control_columns)
            skip_columns.extend(self.conditional_columns)
            skip_columns.extend(self.static_categorical_columns)

            self.target_columns = [c for c in dataset.columns.to_list() if c not in skip_columns]

    def _estimate_frequency(self, df: pd.DataFrame):
        if self.timestamp_column:
            if self.id_columns:
                grps = df.groupby(self.id_columns)
                _, df_subset = next(iter(grps))
            else:
                df_subset = df

            self.freq = estimate_frequency(df_subset[self.timestamp_column])

            if not isinstance(self.freq, (str, int)):
                self.freq = str(self.freq)

        else:
            # no timestamp, assume sequential count?
            self.freq = 1

    def train(
        self,
        dataset: Union[Dataset, pd.DataFrame],
    ) -> "TimeSeriesPreprocessor":
        """Train data transformation operations

        Currently iterates over groups defined by id_columns to train the scaler, if enabled.
        This could be generalized to arbitrary sequence of operations to apply to each group.

        The fitted scalers and their parameters are saved in scaler_dict

        Returns: self

        """

        self._check_dataset(dataset)
        df = self._standardize_dataframe(dataset)
        self._set_targets(df)
        self._validate_columns()

        if self.freq is None:
            self._estimate_frequency(df)

        if self.encode_categorical:
            self._train_categorical_encoder(df)
            if self.scale_categorical_columns:
                # process now so we can learn the scaling factors
                df = self._process_encoding(df.copy())

        if self.scaling:
            self._train_scaler(df)

        self._clean_up_dataframe(df)
        return self

    def inverse_scale_targets(
        self, dataset: Union[Dataset, pd.DataFrame], suffix: Optional[str] = None
    ) -> Union[Dataset, pd.DataFrame]:
        self._check_dataset(dataset)
        df = self._standardize_dataframe(dataset)

        if not self.scaling:
            return dataset

        if len(self.target_scaler_dict) == 0:
            # trying to inverse scale but this preprocessor is not set up for scaling
            raise RuntimeError(
                "Attempt to perform inverse scaling, but time series preprocessor has not yet been trained. Please run the `train` method first."
            )

        cols_to_scale = self.target_columns
        if suffix is not None:
            cols_to_scale = [f"{c}{suffix}" for c in cols_to_scale]

        col_has_list = [df[c].dtype == np.dtype("O") for c in cols_to_scale]

        def explode_row(df_row, name, columns):
            df = pd.DataFrame(df_row[columns].to_dict())
            inv_scale = self.target_scaler_dict[name].inverse_transform(df)
            df_out = df_row.copy()
            for idx, c in enumerate(columns):
                df_out[c] = inv_scale[:, idx]
            return df_out

        def inverse_scale_func(grp, id_columns):
            if isinstance(id_columns, list):
                name = tuple(grp.iloc[0][id_columns].tolist())
            else:
                name = grp.iloc[0][id_columns]

            if not np.any(col_has_list):
                grp[cols_to_scale] = self.target_scaler_dict[name].inverse_transform(grp[cols_to_scale])
            else:
                grp[cols_to_scale] = grp[cols_to_scale].apply(
                    lambda x: explode_row(x, name, cols_to_scale), axis="columns"
                )
            return grp

        if self.scaling_id_columns is not None and len(self.scaling_id_columns) > 0:
            id_columns = self.scaling_id_columns if len(self.scaling_id_columns) > 1 else self.scaling_id_columns[0]
        else:
            id_columns = INTERNAL_ID_COLUMN

        df_inv = df.groupby(id_columns, group_keys=False)[df.columns].apply(
            inverse_scale_func,
            id_columns=id_columns,
        )
        self._clean_up_dataframe(df_inv)
        return df_inv

    def preprocess(
        self,
        dataset: Union[Dataset, pd.DataFrame],
    ) -> Dataset:
        """Main function used to return preprocessed data"""
        # for now we assume data is already fully loaded
        # eventually we need a strategy for dealing with:
        # 1) lists of references to datasets
        # 2) incremental / batch based processing of datasets to minimize memory impact

        self._check_dataset(dataset)
        df = self._standardize_dataframe(dataset)

        df = self._process_encoding(df)

        if self.scaling:
            other_cols_to_scale = self._get_other_columns_to_scale()

            if self.scaling and len(self.target_scaler_dict) == 0:
                # trying to get output, but we never trained the scaler
                raise RuntimeError(
                    "Attempt to get scaled output, but scaler has not yet been trained. Please run the `train` method first."
                )

            # note, we might want an option to return a copy of the data rather than modifying in place

            def scale_func(grp, id_columns):
                if isinstance(id_columns, list):
                    name = tuple(grp.iloc[0][id_columns].tolist())
                else:
                    name = grp.iloc[0][id_columns]
                grp[self.target_columns] = self.target_scaler_dict[name].transform(grp[self.target_columns])
                if other_cols_to_scale:
                    grp[other_cols_to_scale] = self.scaler_dict[name].transform(grp[other_cols_to_scale])

                return grp

            if self.scaling_id_columns is not None and len(self.scaling_id_columns) > 0:
                id_columns = (
                    self.scaling_id_columns if len(self.scaling_id_columns) > 1 else self.scaling_id_columns[0]
                )
            else:
                id_columns = INTERNAL_ID_COLUMN

            df_out = df.groupby(id_columns, group_keys=False)[df.columns].apply(
                scale_func,
                id_columns=id_columns,
            )
            df = df_out

        self._clean_up_dataframe(df)
        return df


def prepare_data_splits(
    data: pd.DataFrame,
    id_columns: List[str] = [],
    context_length: int = 64,
    split_config: Dict[str, Union[List[Union[int, float]], float]] = {"train": 0.7, "test": 0.2},
) -> Tuple[pd.DataFrame]:
    """Splits the input dataframe according to the split_config.

    Args:
        data (pd.DataFrame): Input dataframe.
        id_columns (List[str]): List of column names which identify different time series in a multi-time series input. Defaults to [].
        context_length (int, optional): Specifies the length of the context windows extracted from the historical data for feeding into
                the model. Defaults to 64.
        split_config (Dict[str, Union[List[Union[int, float]], float]]): Dictionary of dictionaries containing
            split parameters.  Defaults to {"train": 0.7, "test": 0.2}. Two configurations are possible:
            1. Specify train/valid/test indices or relative fractions
                {
                    train: [0, 50],
                    valid: [50, 70],
                    test:  [70, 100]
                }
            end value is not inclusive
            2. Specify train/test fractions:
                {
                    train: 0.7
                    test: 0.2
                }
                A valid split should not be specified directly; the above implies valid = 0.1
    Returns:
        Tuple of pandas dataframes, including: train, validation, test.
    """
    # get split_params
    split_params, split_function = get_split_params(split_config, context_length=context_length)

    # split data
    if isinstance(split_function, dict):
        train_data = split_function["train"](data, id_columns=id_columns, **split_params["train"])
        valid_data = split_function["valid"](data, id_columns=id_columns, **split_params["valid"])
        test_data = split_function["test"](data, id_columns=id_columns, **split_params["test"])
    else:
        train_data, valid_data, test_data = split_function(data, id_columns=id_columns, **split_params)

    return train_data, valid_data, test_data


def get_datasets(
    ts_preprocessor: TimeSeriesPreprocessor,
    dataset: Union[Dataset, pd.DataFrame],
    split_config: Dict[str, Union[List[Union[int, float]], float]] = {"train": 0.7, "test": 0.2},
    stride: int = 1,
    fewshot_fraction: Optional[float] = None,
    fewshot_location: str = FractionLocation.LAST.value,
    as_univariate: bool = False,
    use_frequency_token: bool = False,
    enable_padding: bool = True,
    seed: int = 42,
    **dataset_kwargs,
) -> Tuple[Any]:
    """Creates the preprocessed pytorch datasets needed for training and evaluation
    using the HuggingFace trainer

    Args:
        dataset (Union[Dataset, pd.DataFrame]): Loaded pandas dataframe
        split_config (Dict[str, Union[List[Union[int, float]], float]]): Dictionary of dictionaries containing
            split parameters. Defaults to {"train": 0.7, "test": 0.2}. Two configurations are possible:
            1. Specify train/valid/test indices or relative fractions
                {
                    train: [0, 50],
                    valid: [50, 70],
                    test:  [70, 100]
                }
            end value is not inclusive
            2. Specify train/test fractions:
                {
                    train: 0.7
                    test: 0.2
                }
                A valid split should not be specified directly; the above implies valid = 0.1
        stride (int): Stride used for creating the datasets. It is applied to all of train, validation, and test.
            Defaults to 1.
        fewshot_fraction (float, optional): When non-null, return this percent of the original training
            dataset. This is done to support fewshot fine-tuning.
        fewshot_location (str): Determines where the fewshot data is chosen. Valid options are "first", "last" and "uniform"
            as described in the enum FewshotLocation. Default is to choose the fewshot data at the end
            of the training dataset (i.e., "last"). If fewshot enabled before windowing, then we support first and last,
            if fewshot enabled after windowing, then we support uniform sampling.
        as_univariate (bool, optional): When True the datasets returned will contain only one target column. An
            additional ID is added to distinguish original column name. Only valid if there are no exogenous
            specified. Defaults to False.
        use_frequency_token (bool): If True, datasets are created that include the frequency token. Defaults to False.
        enable_padding (bool): If True, datasets are created with padding. Padding will add zeros to the dataframe (per
            time series) when there is insufficient data to form one record. If False, no padding is done and one or
            more datasets may be empty.
        seed (int): Seed to use.
        dataset_kwargs: additional keyword arguments to pass to the torch datasets during creation.


    Returns:
        Tuple of pytorch datasets, including: train, validation, test.
    """

    rng = np.random.default_rng(seed=seed)

    if ts_preprocessor.context_length is None:
        raise ValueError("TimeSeriesPreprocessor must be instantiated with non-null context_length")
    if ts_preprocessor.prediction_length is None:
        raise ValueError("TimeSeriesPreprocessor must be instantiated with non-null prediction_length")

    data = ts_preprocessor._standardize_dataframe(dataset)

    train_data, valid_data, test_data = prepare_data_splits(
        data,
        id_columns=ts_preprocessor.id_columns,
        context_length=ts_preprocessor.context_length,
        split_config=split_config,
    )

    # data preprocessing
    ts_preprocessor.train(train_data)

    # specify columns
    column_specifiers = {
        "id_columns": ts_preprocessor.id_columns,
        "timestamp_column": ts_preprocessor.timestamp_column,
        "target_columns": ts_preprocessor.target_columns,
        "observable_columns": ts_preprocessor.observable_columns,
        "control_columns": ts_preprocessor.control_columns,
        "conditional_columns": ts_preprocessor.conditional_columns,
        "categorical_columns": ts_preprocessor.categorical_columns,
        "static_categorical_columns": ts_preprocessor.static_categorical_columns,
    }

    # handle fewshot operation
    if (fewshot_fraction is not None) and not ((fewshot_fraction <= 1.0) and (fewshot_fraction > 0.0)):
        raise ValueError(f"Fewshot fraction should be between 0 and 1, received {fewshot_fraction}")
    if fewshot_fraction is not None and fewshot_location != FractionLocation.UNIFORM.value:
        train_data = select_by_fixed_fraction(
            train_data,
            id_columns=ts_preprocessor.id_columns,
            fraction=fewshot_fraction,
            location=fewshot_location,
            minimum_size=ts_preprocessor.context_length,
        )

    params = column_specifiers
    params["context_length"] = ts_preprocessor.context_length
    params["prediction_length"] = ts_preprocessor.prediction_length
    params["stride"] = stride
    params["enable_padding"] = enable_padding
    if use_frequency_token:
        params["frequency_token"] = ts_preprocessor.get_frequency_token(ts_preprocessor.freq)

    # get torch datasets
    train_valid_test = [train_data, valid_data, test_data]
    train_valid_test_prep = [ts_preprocessor.preprocess(d) for d in train_valid_test]

    if as_univariate and len(ts_preprocessor.target_columns) > 1:
        if (
            ts_preprocessor.observable_columns
            or ts_preprocessor.control_columns
            or ts_preprocessor.conditional_columns
            or ts_preprocessor.static_categorical_columns
        ):
            raise ValueError("`as_univariate` option only allowed when there are no exogenous columns.")

        train_valid_test_prep = [
            convert_to_univariate(
                d,
                timestamp_column=ts_preprocessor.timestamp_column,
                id_columns=ts_preprocessor.id_columns,
                target_columns=ts_preprocessor.target_columns,
            )
            for d in train_valid_test_prep
        ]

        params["target_columns"] = ["value"]
        params["id_columns"] = params["id_columns"] + ["column_id"]

    params.update(**dataset_kwargs)

    datasets = tuple([ForecastDFDataset(d, **params) for d in train_valid_test_prep])
    for dset_name, dset in zip(["train", "valid", "test"], datasets):
        if len(dset) == 0:
            raise RuntimeError(f"One of the generated datasets ({dset_name}) is of zero length.")

    if fewshot_fraction is not None and fewshot_location == FractionLocation.UNIFORM.value:
        lst = rng.integers(low=0, high=len(datasets[0]), size=int(fewshot_fraction * len(datasets[0])))
        few_shot_train = Subset(datasets[0], lst.tolist())

        datasets = (few_shot_train, datasets[1], datasets[2])

    return datasets


def create_timestamps(
    last_timestamp: Union[datetime.datetime, pd.Timestamp, np.datetime64, int, float, np.integer, np.floating],
    freq: Optional[
        Union[int, float, np.integer, np.floating, datetime.timedelta, pd.Timedelta, np.timedelta64, str]
    ] = None,
    time_sequence: Optional[Union[List[int], List[float], List[datetime.datetime], List[pd.Timestamp]]] = None,
    periods: int = 1,
) -> List[pd.Timestamp]:
    """Simple utility to create a list of timestamps based on start, delta and number of periods

    Args:
        last_timestamp (Union[datetime.datetime, pd.Timestamp, int, float, np.integer, np.floating]): The last observed timestamp, new timestamps will be created
            after this timestamp.
        freq (Optional[Union[int, float, np.integer, np.floating, datetime.timedelta, pd.Timedelta, np.timedelta, str]], optional): The frequency at which timestamps
            should be generated. Defaults to None.
        time_sequence (Optional[Union[List[int], List[float], List[datetime.datetime], List[pd.Timestamp]]], optional): A time sequence
            from which the frequency can be inferred. Defaults to None.
        periods (int, optional): The number of timestamps to generate. Defaults to 1.

    Raises:
        ValueError: If the frequency cannot be parsed from freq or inferred from time_sequence

    Returns:
        List[pd.Timestamp]: List of timestamps
    """

    if freq is None and time_sequence is None:
        raise ValueError("Neither `freq` nor `time_sequence` provided, cannot determine frequency.")

    if freq is None:
        freq = estimate_frequency(time_sequence)

    if freq is None:
        raise ValueError(
            "Could not extend time series because frequency was not provided and could not be estimated from the available data."
        )

    def convert_numeric(val: Any):
        """Helper function to convert strings to numerical values"""
        if isinstance(val, str):
            try:
                return int(val)
            except (ValueError, TypeError):
                try:
                    return float(val)
                except (ValueError, TypeError):
                    return val
        return val

    if isinstance(freq, str):
        freq = convert_numeric(freq)

    # freq is str, which is not convertible to numeric
    if isinstance(freq, str):
        # try to convert to pd.timedelta
        try:
            freq = pd._libs.tslibs.timedeltas.Timedelta(freq)
        except ValueError:
            pass

        return pd.date_range(
            last_timestamp,
            freq=freq,
            periods=periods + 1,
        ).tolist()[1:]
    # frequency is timedelta like object
    elif isinstance(freq, (pd.Timedelta, np.timedelta64, datetime.timedelta)):
        if isinstance(last_timestamp, (np.datetime64, pd.Timestamp, datetime.datetime)):
            return pd.date_range(
                last_timestamp,
                freq=pd.Timedelta(freq) if isinstance(freq, np.timedelta64) else freq,
                periods=periods + 1,
            ).tolist()[1:]
            # [last_timestamp + i * freq for i in range(1, periods + 1)]
        # last_timestamp is not date type, but freq is timedelta type -- ambiguous
        else:
            raise ValueError(
                f"Ambiguous last_timestamp {last_timestamp} (type: {type(last_timestamp)}) with freq {freq}."
            )
    # frequency is numeric
    elif isinstance(freq, (int, float, np.integer, np.floating)):
        if isinstance(last_timestamp, (int, float, np.floating, np.integer)):
            return [last_timestamp + i * freq for i in range(1, periods + 1)]
        # last_timestamp is a date type, but freq is numeric -- ambiguous
        else:
            raise ValueError(
                f"Ambiguous frequency {freq} with last_timestamp {last_timestamp} (type: {type(last_timestamp)})."
            )
    else:
        raise ValueError(
            f"Could not create timestamps, given the following inputs: last_timestamp={last_timestamp}, freq={freq}, periods={periods}."
        )


def estimate_frequency(timestamp_data: Union[pd.Series, np.ndarray]):
    if len(timestamp_data) < 2:
        LOGGER.warning("Provided time series data is too short to estimate frequency.")
        return None

    if isinstance(timestamp_data, pd.Series):
        return timestamp_data.iloc[-1] - timestamp_data.iloc[-2]
    else:
        return timestamp_data[-1] - timestamp_data[-2]


def extend_time_series(
    time_series: pd.DataFrame,
    timestamp_column: str,
    grouping_columns: List[str],
    freq: Optional[Union[int, float, datetime.timedelta, pd.Timedelta]] = None,
    periods: int = None,
    total_periods: Optional[int] = None,
):
    """Extends the provided time series with empty data for the number of periods specified. For each time series, based
    on groups defined by grouping columns, adds emptry records following the last timestamp. The empty records contain
    only timestamps and grouping indicators, remaining fields will be null.

    One of periods or total_periods must be specified.

    Args:
        time_series (pd.DataFrame): _description_
        start_timestamp (_type_): _description_
        column_name (str): _description_
        grouping_columns (List[str]): _description_
        freq:
        periods (int, optional): _description_. Defaults to 1.
        total_periods (int, optional): total length of the series after extending. Defaults to None.
    """

    def augment_one_series(
        group: Union[pd.Series, pd.DataFrame], periods: Optional[int] = None, total_periods: Optional[int] = None
    ):
        last_timestamp = group[timestamp_column].iloc[-1]

        if periods is None:
            periods = total_periods - len(group)

        if periods < 1:
            return group

        new_data = pd.DataFrame(
            {
                timestamp_column: create_timestamps(
                    last_timestamp,
                    freq=freq,
                    time_sequence=group[timestamp_column].values,
                    periods=periods,
                )
            }
        )

        df = pd.concat((group, new_data), axis=0, ignore_index=True)
        return df  # df.reset_index(drop=True)

    if (periods is None and total_periods is None) or (periods is not None and total_periods is not None):
        raise ValueError("Exactly one of `periods` or `total_periods` must be specified")

    if grouping_columns == []:
        new_time_series = augment_one_series(time_series, periods=periods, total_periods=total_periods)
    else:
        new_time_series = time_series.groupby(grouping_columns).apply(
            augment_one_series, include_groups=False, periods=periods, total_periods=total_periods
        )
        idx_names = list(new_time_series.index.names)
        idx_names[-1] = "__delete"
        new_time_series = new_time_series.reset_index(names=idx_names)
        new_time_series.drop(columns=["__delete"], inplace=True)

    return new_time_series



# Copyright contributors to the TSFM project
#

from .callbacks import TrackingCallback
from .data_handling import load_dataset
from .dataset import ClassificationDFDataset, ForecastDFDataset, PretrainDFDataset, RegressionDFDataset
from .get_model import get_model
from .lr_finder import optimal_lr_finder
from .recursive_predictor import RecursivePredictor, RecursivePredictorConfig, RecursivePredictorOutput
from .time_series_anomaly_detection_pipeline import TimeSeriesAnomalyDetectionPipeline
from .time_series_classification_pipeline import TimeSeriesClassificationPipeline
from .time_series_classification_preprocessor import TimeSeriesClassificationPreprocessor
from .time_series_forecasting_pipeline import TimeSeriesForecastingPipeline
from .time_series_imputation_pipeline import TimeSeriesImputationPipeline
from .time_series_preprocessor import TimeSeriesPreprocessor, get_datasets
from .util import count_parameters
